from django.db import migrations


def operation_copy_entries(apps, schema_editor):
    SharedUploadedFile = apps.get_model(
        app_label='common', model_name='SharedUploadedFile'
    )
    StorageSharedUploadedFile = apps.get_model(
        app_label='storage', model_name='StorageSharedUploadedFile'
    )

    for entry in SharedUploadedFile.objects.using(schema_editor.connection.alias).all():
        StorageSharedUploadedFile.objects.create(
            file=entry.file, filename=entry.filename,
            datetime=entry.datetime
        )
        entry.delete()


def operation_copy_entries_reverse(apps, schema_editor):
    SharedUploadedFile = apps.get_model(
        app_label='common', model_name='SharedUploadedFile'
    )
    StorageSharedUploadedFile = apps.get_model(
        app_label='storage', model_name='StorageSharedUploadedFile'
    )

    for entry in StorageSharedUploadedFile.objects.using(schema_editor.connection.alias).all():
        SharedUploadedFile.objects.create(
            file=entry.file, filename=entry.filename,
            datetime=entry.datetime
        )
        entry.delete()


class Migration(migrations.Migration):
    dependencies = [
        ('storage', '0001_initial'), ('common', '0015_auto_20200501_0631')
    ]

    operations = [
        migrations.RunPython(
            code=operation_copy_entries,
            reverse_code=operation_copy_entries_reverse
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_type_web_links, icon_document_web_link_list,
    icon_web_link_create, icon_web_link_delete, icon_web_link_document_types,
    icon_web_link_edit, icon_web_link_instance_view, icon_web_link_list,
    icon_web_link_setup
)
from .permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_instance_view,
)

link_document_type_web_links = Link(
    args='resolved_object.pk', icon=icon_document_type_web_links,
    permissions=(permission_document_type_edit,), text=_('Web links'),
    view='web_links:document_type_web_links',
)
link_document_web_link_list = Link(
    args='resolved_object.pk', icon=icon_document_web_link_list,
    permissions=(permission_web_link_instance_view,), text=_('Web links'),
    view='web_links:document_web_link_list',
)
link_web_link_create = Link(
    icon=icon_web_link_create,
    permissions=(permission_web_link_create,), text=_('Create new web link'),
    view='web_links:web_link_create'
)
link_web_link_delete = Link(
    args='object.pk', icon=icon_web_link_delete,
    permissions=(permission_web_link_delete,),
    tags='dangerous', text=_('Delete'), view='web_links:web_link_delete',
)
link_web_link_document_types = Link(
    args='object.pk', icon=icon_web_link_document_types,
    permissions=(permission_web_link_edit,), text=_('Document types'),
    view='web_links:web_link_document_types',
)
link_web_link_edit = Link(
    args='object.pk', icon=icon_web_link_edit,
    permissions=(permission_web_link_edit,), text=_('Edit'),
    view='web_links:web_link_edit',
)
link_web_link_instance_view = Link(
    icon=icon_web_link_instance_view,
    args=('document.pk', 'object.pk',),
    permissions=(permission_web_link_instance_view,), tags='new_window',
    text=_('Navigate'), view='web_links:web_link_instance_view',
)
link_web_link_list = Link(
    icon=icon_web_link_list, text=_('Web links'),
    view='web_links:web_link_list'
)
link_web_link_setup = Link(
    icon=icon_web_link_setup,
    permissions=(permission_web_link_create,), text=_('Web links'),
    view='web_links:web_link_list'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Messaging'), name='messaging')

event_message_created = namespace.add_event_type(
    label=_('Message created'), name='message_created'
)
event_message_edited = namespace.add_event_type(
    label=_('Message edited'), name='message_edited'
)



from mayan.apps.authentication.tests.mixins import LoginViewTestMixin, LogoutViewTestMixin
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.user_management.permissions import (
    permission_user_edit, permission_user_view
)

from ..events import event_user_locale_profile_edited

from .mixins import UserLocaleProfileViewMixin
from .literals import TEST_TRANSLATED_WORD


class CurrentUserViewTestCase(
    UserLocaleProfileViewMixin, GenericViewTestCase
):
    def test_current_user_locale_profile_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_current_user_locale_profile_detail_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_locale_profile_edit_view_no_permission(self):
        language = self._test_case_user.locale_profile.language
        timezone = self._test_case_user.locale_profile.timezone

        self._clear_events()

        response = self._request_test_current_user_locale_profile_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()
        self.assertNotEqual(
            self._test_case_user.locale_profile.language, language
        )
        self.assertNotEqual(
            self._test_case_user.locale_profile.timezone, timezone
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_locale_profile_edited.id)


class SuperUserLocaleViewTestCase(
    UserLocaleProfileViewMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_superuser()

    def test_superuser_locale_profile_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_superuser_locale_profile_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_superuser_locale_profile_detail_view_with_access(self):
        self.grant_access(
            obj=self.test_superuser, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_superuser_locale_profile_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_superuser_locale_profile_edit_view_no_permission(self):
        language = self.test_superuser.locale_profile.language
        timezone = self.test_superuser.locale_profile.timezone

        self._clear_events()

        response = self._request_test_superuser_locale_profile_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_superuser.refresh_from_db()
        self.assertEqual(
            self.test_superuser.locale_profile.language, language
        )
        self.assertEqual(
            self.test_superuser.locale_profile.timezone, timezone
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_superuser_locale_profile_edit_view_with_access(self):
        language = self.test_superuser.locale_profile.language
        timezone = self.test_superuser.locale_profile.timezone

        self.grant_access(
            obj=self.test_superuser, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_superuser_locale_profile_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_superuser.refresh_from_db()
        self.assertEqual(
            self.test_superuser.locale_profile.language, language
        )
        self.assertEqual(
            self.test_superuser.locale_profile.timezone, timezone
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class UserLocaleProfileViewTestCase(
    UserLocaleProfileViewMixin, GenericViewTestCase
):
    auto_create_test_user = True

    def test_user_locale_profile_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_user_locale_profile_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_locale_profile_detail_view_with_access(self):
        self.grant_access(
            obj=self.test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_locale_profile_detail_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_locale_profile_edit_view_no_permission(self):
        language = self.test_user.locale_profile.language
        timezone = self.test_user.locale_profile.timezone

        self._clear_events()

        response = self._request_test_user_locale_profile_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_user.refresh_from_db()
        self.assertEqual(self.test_user.locale_profile.language, language)
        self.assertEqual(self.test_user.locale_profile.timezone, timezone)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_locale_profile_edit_view_with_access(self):
        language = self.test_user.locale_profile.language
        timezone = self.test_user.locale_profile.timezone

        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_locale_profile_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_user.refresh_from_db()
        self.assertNotEqual(self.test_user.locale_profile.language, language)
        self.assertNotEqual(self.test_user.locale_profile.timezone, timezone)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_locale_profile_edited.id)


class LanguageSelectionViewTestCase(
    LoginViewTestMixin, LogoutViewTestMixin, UserLocaleProfileViewMixin,
    GenericViewTestCase
):
    def test_language_change_view(self):
        response = self._request_test_current_user_locale_profile_edit_view(
            follow=True
        )
        self.assertEqual(response.status_code, 200)

        response = self._request_test_current_user_locale_profile_detail_view()
        self.assertContains(
            response=response, text=TEST_TRANSLATED_WORD, status_code=200
        )

    def test_language_change_after_login_view(self):
        response = self._request_test_current_user_locale_profile_edit_view(
            follow=True
        )
        self.assertEqual(response.status_code, 200)

        request = self._request_logout_view()
        self.assertEqual(request.status_code, 302)

        response = self._request_simple_login_view()
        self.assertEqual(request.status_code, 302)

        response = self._request_test_current_user_locale_profile_detail_view()
        self.assertContains(
            response=response, text=TEST_TRANSLATED_WORD, status_code=200
        )



from rest_framework import serializers as rest_framework_serializers
from rest_framework.fields import (  # NOQA
    BooleanField, CharField, ChoiceField, DateField, DateTimeField,
    DecimalField, DictField, DurationField, EmailField, Field, FileField,
    FilePathField, FloatField, HiddenField, HStoreField, IPAddressField,
    ImageField, IntegerField, JSONField, ListField, ModelField,
    MultipleChoiceField, NullBooleanField, ReadOnlyField, RegexField,
    SerializerMethodField, SlugField, TimeField, URLField, UUIDField
)
from rest_framework.serializers import (
    HyperlinkedModelSerializer as RESTFrameworkHyperlinkedModelSerializer,
    ModelSerializer as RESTFrameworkModelSerializer
)
from rest_framework.relations import (  # NOQA
    HyperlinkedIdentityField, HyperlinkedRelatedField, ManyRelatedField,
    PrimaryKeyRelatedField, RelatedField, SlugRelatedField,
    StringRelatedField
)
from rest_framework.reverse import reverse

from .classes import BatchRequestCollection
from .serializer_mixins import DynamicFieldListSerializerMixin


class Serializer(
    DynamicFieldListSerializerMixin, rest_framework_serializers.Serializer
):
    """Serializer subclass to add Mayan specific mixins."""


class BatchAPIRequestResponseSerializer(Serializer):
    content = CharField(read_only=True)
    data = JSONField(read_only=True)
    headers = DictField(read_only=True)
    name = CharField(read_only=True)
    status_code = IntegerField(read_only=True)
    requests = JSONField(
        style={'base_template': 'textarea.html'},
        write_only=True
    )

    def validate(self, data):
        try:
            BatchRequestCollection(request_list=data['requests'])
        except Exception as exception:
            raise rest_framework_serializers.ValidationError(
                'Error validating requests; {}'.format(exception)
            )

        return data


class BlankSerializer(Serializer):
    """Serializer for the object action API view."""


class EndpointSerializer(Serializer):
    label = CharField(read_only=True)
    url = SerializerMethodField()

    def get_url(self, instance):
        if instance.viewname:
            return reverse(
                kwargs=instance.kwargs, viewname=instance.viewname,
                request=self.context['request'],
                format=self.context['format']
            )


class HyperlinkedModelSerializer(
    DynamicFieldListSerializerMixin, RESTFrameworkHyperlinkedModelSerializer
):
    """HyperlinkedModelSerializer subclass to add Mayan specific mixins."""


class ModelSerializer(
    DynamicFieldListSerializerMixin, RESTFrameworkModelSerializer
):
    """ModelSerializer subclass to add Mayan specific mixins."""


class ProjectInformationSerializer(Serializer):
    __title__ = CharField(read_only=True)
    __version__ = CharField(read_only=True)
    __build__ = CharField(read_only=True)
    __build_string__ = CharField(read_only=True)
    __django_version__ = CharField(read_only=True)
    __author__ = CharField(read_only=True)
    __author_email__ = CharField(read_only=True)
    __description__ = CharField(read_only=True)
    __license__ = CharField(read_only=True)
    __copyright__ = CharField(read_only=True)
    __website__ = CharField(read_only=True)



from django import forms
from django.core.exceptions import ImproperlyConfigured

from mayan.apps.acls.models import AccessControlList


class FilteredModelFieldMixin:
    def __init__(self, *args, **kwargs):
        self.source_model = kwargs.pop('source_model', None)
        self.permission = kwargs.pop('permission', None)
        self.source_queryset = kwargs.pop('source_queryset', None)

        if self.source_queryset is None:
            if self.source_model:
                self.source_queryset = self.source_model._meta.default_manager.all()
            else:
                raise ImproperlyConfigured(
                    '{} requires a source_queryset or a source_model to be '
                    'specified as keyword argument.'.format(
                        self.__class__.__name__
                    )
                )

        kwargs['queryset'] = self.source_queryset.none()

        super().__init__(*args, **kwargs)

    def reload(self):
        if self.permission and self.user:
            self.queryset = AccessControlList.objects.restrict_queryset(
                permission=self.permission, queryset=self.source_queryset,
                user=self.user
            )
        else:
            self.queryset = self.source_queryset


class FilteredModelChoiceField(
    FilteredModelFieldMixin, forms.ModelChoiceField
):
    """Single selection filtered model choice field"""


class FilteredModelMultipleChoiceField(
    FilteredModelFieldMixin, forms.ModelMultipleChoiceField
):
    """Multiple selection filtered model choice field"""



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('checkouts', '0008_checkedoutdocument')
    ]

    operations = [
        migrations.RenameField(
            model_name='documentcheckout',
            old_name='block_new_version',
            new_name='block_new_file',
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.permissions import Permission
from mayan.apps.permissions.serializers import RoleSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)
from mayan.apps.rest_api.serializer_mixins import CreateOnlyFieldSerializerMixin

from .models import AccessControlList


class ACLSerializer(
    CreateOnlyFieldSerializerMixin, serializers.ModelSerializer
):
    content_type = ContentTypeSerializer(read_only=True)
    permissions_add_url = MultiKwargHyperlinkedIdentityField(
        view_name='rest_api:accesscontrollist-permission-add',
        view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label',
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name',
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id',
            }
        ),
    )
    permissions_remove_url = MultiKwargHyperlinkedIdentityField(
        view_name='rest_api:accesscontrollist-permission-remove',
        view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label',
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name',
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id',
            }
        ),
    )
    permissions_url = MultiKwargHyperlinkedIdentityField(
        view_name='rest_api:accesscontrollist-permission-list',
        view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label',
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name',
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id',
            }
        ),
    )
    role = RoleSerializer(read_only=True)
    role_id = serializers.IntegerField(write_only=True)
    url = MultiKwargHyperlinkedIdentityField(
        view_name='rest_api:accesscontrollist-detail',
        view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label',
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name',
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id',
            }
        ),
    )

    class Meta:
        create_only_fields = ('role_id',)
        fields = (
            'content_type', 'id', 'object_id', 'permissions_add_url',
            'permissions_remove_url', 'permissions_url', 'role', 'role_id',
            'url'
        )
        model = AccessControlList
        read_only_fields = ('content_type', 'object_id')


class ACLPermissionAddSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_('Primary key of the permission to add to the ACL.'),
        source_queryset=Permission.all()
    )


class ACLPermissionRemoveSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_('Primary key of the permission to remove from the ACL.'),
        source_queryset=Permission.all()
    )



from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.testing.tests.utils import mute_stdout

from ..models import AutoAdminSingleton

from .literals import TEST_FIRST_TIME_LOGIN_TEXT, TEST_MOCK_VIEW_TEXT
from .mixins import AutoAdminViewMixing


class AutoAdminViewCase(AutoAdminViewMixing, GenericViewTestCase):
    auto_create_group = False
    auto_create_users = False
    auto_login_user = False

    def setUp(self):
        super().setUp()
        with mute_stdout():
            AutoAdminSingleton.objects.create_autoadmin()

    def test_login_302_view(self):
        response = self._request_home_view()

        self.assertContains(
            response=response, text=TEST_FIRST_TIME_LOGIN_TEXT,
            status_code=200
        )

    def test_login_ok_view(self):
        autoadmin = AutoAdminSingleton.objects.get()
        logged_in = self.login(
            username=autoadmin.account,
            password=autoadmin.password
        )
        self.assertTrue(logged_in)

        response = self._request_home_view()

        self.assertNotContains(
            response=response, text=TEST_MOCK_VIEW_TEXT,
            status_code=200
        )



from mayan.apps.rest_api import serializers

from .models import Asset


class AssetSerializer(serializers.HyperlinkedModelSerializer):
    image_url = serializers.HyperlinkedIdentityField(
        lookup_url_kwarg='asset_id',
        view_name='rest_api:asset-image'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_url_kwarg': 'asset_id',
                'view_name': 'rest_api:asset-detail'
            },
        }
        fields = (
            'file', 'label', 'id', 'image_url', 'internal_name', 'url'
        )
        model = Asset



import json
from io import BytesIO
import logging
from packaging import version
from pathlib import Path
import pkg_resources
import shutil
import sys
import tarfile

from furl import furl
import requests
from semver import max_satisfying

from django.apps import apps
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.termcolors import colorize
from django.utils.translation import ugettext_lazy as _, ugettext

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.exceptions import ResolverPipelineError
from mayan.apps.common.utils import ResolverPipelineObjectAttribute
from mayan.apps.storage.utils import (
    TemporaryDirectory, mkdtemp, patch_files as storage_patch_files
)

from .algorithms import HashAlgorithm
from .environments import environment_production
from .exceptions import DependenciesException

logger = logging.getLogger(name=__name__)


class Provider:
    """Base provider class"""


class PyPIRespository(Provider):
    url = 'https://pypi.org/'


class GoogleFontsProvider(Provider):
    url = 'https://fonts.googleapis.com/'


class NPMRegistryRespository(Provider):
    url = 'http://registry.npmjs.com'


class OperatingSystemProvider(Provider):
    """Placeholder for the OS provider"""


class DependencyGroup:
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_all(cls):
        return sorted(cls._registry.values(), key=lambda x: x.label)

    def __init__(
        self, attribute_name, label, name, allow_multiple=False,
        help_text=None
    ):
        self.allow_multiple = allow_multiple
        self.attribute_name = attribute_name
        self.label = label
        self.help_text = help_text
        self.name = name

        self.__class__._registry[name] = self

    def __str__(self):
        return force_text(s=self.label)

    @staticmethod
    def get_options_for_dependency_group(dependency_group):
        result = []

        for dependency in Dependency.get_all():
            value = ResolverPipelineObjectAttribute.resolve(
                attribute=dependency_group.attribute_name, obj=dependency
            )

            try:
                label = ResolverPipelineObjectAttribute.resolve(
                    attribute='{}_verbose_name'.format(
                        dependency_group.attribute_name
                    ), obj=dependency
                )
            except ResolverPipelineError:
                label = value

            try:
                help_text = ResolverPipelineObjectAttribute.resolve(
                    attribute='{}_help_text'.format(
                        dependency_group.attribute_name
                    ), obj=dependency
                )
            except ResolverPipelineError:
                if dependency_group.allow_multiple:
                    help_text = (None,) * len(value)
                else:
                    help_text = None

            if dependency_group.allow_multiple:
                for entry_index, entry in enumerate(value):
                    dictionary = {
                        'label': label[entry_index], 'help_text': help_text[entry_index], 'value': entry
                    }
                    if dictionary not in result:
                        result.append(dictionary)
            else:
                dictionary = {
                    'label': label, 'help_text': help_text, 'value': value
                }
                if dictionary not in result:
                    result.append(dictionary)

        return result

    def get_entries(self):
        options = DependencyGroup.get_options_for_dependency_group(
            dependency_group=self
        )
        result = []

        for option in options:
            result.append(
                DependencyGroupEntry(
                    dependency_group=self, help_text=option['help_text'],
                    label=option['label'], name=option['value']
                )
            )

        return sorted(result, key=lambda x: x.label)

    def get_entry(self, entry_name):
        for entry in self.get_entries():
            if entry.name == entry_name:
                return entry

        raise KeyError('Entry not found.')


class DependencyGroupEntry:
    def __init__(
        self, dependency_group, label, name, help_text=None
    ):
        self.dependency_group = dependency_group
        self.help_text = help_text or ''
        self.label = label
        self.name = name

    def __str__(self):
        return force_text(s=self.label)

    def get_dependencies(self):
        dependencies = Dependency.get_for_attribute(
            attribute_name=self.dependency_group.attribute_name,
            attribute_value=self.name
        )

        return Dependency.return_sorted(dependencies=dependencies)


class Dependency(AppsModuleLoaderMixin):
    _loader_module_name = 'dependencies'
    _registry = {}

    @staticmethod
    def return_sorted(dependencies):
        return sorted(dependencies, key=lambda x: x.get_label())

    @classmethod
    def _check_all(cls):
        result = []

        for dependency in cls.get_all():
            check = dependency.check()

            if check and any(map(lambda x: x.mark_missing, dependency.environments)):
                check_text = '* {} *'.format(check)
                check_color = check if check else colorize(
                    text=check_text, fg='red', opts=(
                        'bold', 'blink', 'reverse'
                    )
                )

                result.append(
                    {
                        'check': check,
                        'check_text': check_text,
                        'check_color': check_color,
                        'dependency': dependency
                    }
                )

        return result

    @classmethod
    def check_all(cls, as_csv=False, use_color=False):
        if as_csv:
            template = '{},{},{},{},{},{},{}'

            print(
                template.format(
                    ugettext('Name'), ugettext('Type'), ugettext('Version'),
                    ugettext('App'), ugettext('Environments'),
                    ugettext('Other data'), ugettext('Check')
                )
            )
            for result in cls._check_all():
                dependency = result['dependency']

                print(
                    template.format(
                        dependency.name,
                        force_text(s=dependency.class_name_verbose_name),
                        force_text(s=dependency.get_version_string()),
                        force_text(s=dependency.app_label_verbose_name()),
                        force_text(s=dependency.get_environments_verbose_name()),
                        force_text(s=dependency.get_other_data()),
                        force_text(s=result['check'])
                    )
                )
        else:
            for result in cls._check_all():
                dependency = result['dependency']
                print('-' * 40)
                print('* {}'.format(dependency.name))
                print(
                    'Class: {class_name} | Version: {version} '
                    '| App: {app_label} | Environments: {environments} '
                    '| Other data: {other} | Check: {check}'.format(
                        class_name=dependency.class_name_verbose_name,
                        version=dependency.get_version_string(),
                        app_label=dependency.app_label_verbose_name(),
                        environments=dependency.get_environments_verbose_name(),
                        other=dependency.get_other_data(),
                        check=result['check_color'] if use_color else result['check_text']
                    )
                )

        sys.stdout.flush()

    @classmethod
    def get(cls, pk):
        return cls._registry[pk]

    @classmethod
    def get_all(cls, subclass_only=False):
        dependencies = cls._registry.values()
        if subclass_only:
            dependencies = [dependency for dependency in dependencies if isinstance(dependency, cls)]

        return Dependency.return_sorted(dependencies=dependencies)

    @classmethod
    def get_for_attribute(cls, attribute_name, attribute_value, **kwargs):
        result = []

        for dependency in cls.get_all(**kwargs):
            resolved_attibute_value = ResolverPipelineObjectAttribute.resolve(
                attribute=attribute_name, obj=dependency
            )

            if attribute_value == resolved_attibute_value or attribute_value in resolved_attibute_value:
                result.append(dependency)

        return result

    @classmethod
    def install_multiple(cls, app_label=None, force=False, subclass_only=False):
        for dependency in cls.get_all(subclass_only=subclass_only):
            if app_label:
                if app_label == dependency.app_label:
                    dependency.install(force=force)
            else:
                dependency.install(force=force)

    def __init__(
        self, name, environment=environment_production, app_label=None, copyright_text=None,
        environments=None, help_text=None, label=None, module=None, replace_list=None,
        version_string=None
    ):
        self._app_label = app_label
        self.copyright_text = copyright_text
        self.environments = environments or (environment,)
        self.help_text = help_text
        self.label = label
        self.module = module
        self.name = name
        self.package_metadata = None
        self.replace_list = replace_list
        self.repository = self.provider_class()
        self.version_string = version_string

        if not app_label:
            if not module:
                raise DependenciesException(
                    _('Need to specify at least one: app_label or module.')
                )

        if self.get_pk() in self.__class__._registry:
            raise DependenciesException(
                _('Dependency "%s" already registered.') % self.name
            )

        self.__class__._registry[self.get_pk()] = self

    @cached_property
    def app_label(self):
        if not self._app_label:
            app = apps.get_containing_app_config(object_name=self.module)
            return app.label
        else:
            return self._app_label

    def app_label_verbose_name(self):
        return apps.get_app_config(app_label=self.app_label).verbose_name

    def download(self):
        """
        Download the dependency from a repository
        """
        raise NotImplementedError

    def get_copyright(self):
        return self.copyright_text or ''

    def install(self, force=False):
        print(_('Installing package: %s... ') % self.get_label_full(), end='')
        sys.stdout.flush()

        if not force:
            if self.check():
                print(_('Already installed.'))
            else:
                self._install()
                print(_('Complete.'))
                sys.stdout.flush()
        else:
            if self.replace_list:
                self.patch_files()
                print(_('Complete.'))
                sys.stdout.flush()

            self.patch_files()
            print(_('Complete.'))
            sys.stdout.flush()

    def _install(self):
        raise NotImplementedError

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    def check(self):
        """
        Returns the version found or an exception
        """
        if self._check():
            return True
        else:
            return False

    def check_string(self):
        if self._check():
            return 'True'
        else:
            return 'False'

    def check_string_verbose_name(self):
        if self._check():
            return _('Installed and correct version')
        else:
            return _('Missing or incorrect version')

    def _check(self):
        raise NotImplementedError

    def get_help_text(self):
        return self.help_text or ''

    def get_environments(self):
        return [
            environment.name for environment in self.environments
        ]

    def get_environments_help_text(self):
        return [
            environment.help_text for environment in self.environments
        ]

    def get_environments_verbose_name(self):
        return [
            environment.label for environment in self.environments
        ]

    def get_label(self):
        return self.label or self.name

    def get_label_full(self):
        if self.version_string:
            version_string = '({})'.format(self.version_string)
        else:
            version_string = ''

        return '{} {}'.format(self.get_label(), version_string)

    def get_other_data(self):
        return _('None')

    def get_pk(self):
        return self.name

    def get_url(self):
        raise NotImplementedError

    def get_version_string(self):
        return self.version_string or _('Not specified')

    def patch_files(self, path=None, replace_list=None):
        print(_('Patching files... '), end='')

        try:
            sys.stdout.flush()
        except AttributeError:
            pass

        if not path:
            path = self.get_install_path()

        if not replace_list:
            replace_list = self.replace_list

        storage_patch_files(path=path, replace_list=replace_list)

    def verify(self):
        """
        Verify the integrity of the dependency
        """
        raise NotImplementedError


# Depedency subclasses


class BinaryDependency(Dependency):
    class_name = 'binary'
    class_name_help_text = _(
        'Executables that are called directly by the code.'
    )
    class_name_verbose_name = _('Binary')
    provider_class = OperatingSystemProvider

    def __init__(self, *args, **kwargs):
        self.path = kwargs.pop('path')
        super().__init__(*args, **kwargs)

    def _check(self):
        return Path(self.path).exists()

    def get_other_data(self):
        return 'Path: {}'.format(self.path)


class JavaScriptDependency(Dependency):
    class_name = 'javascript'
    class_name_help_text = _(
        'JavaScript libraries downloaded the from NPM registry and used for '
        'front-end functionality.'
    )
    class_name_verbose_name = _('JavaScript')
    provider_class = NPMRegistryRespository

    def __init__(self, *args, **kwargs):
        self.static_folder = kwargs.pop('static_folder', None)
        return super().__init__(*args, **kwargs)

    def _check(self):
        try:
            package_info = self._read_package_file()
        except FileNotFoundError:
            return False

        versions = [package_info['version']]
        version_string = self.version_string

        return max_satisfying(
            versions=versions, range_=version_string,
            loose=True
        )

    def _read_package_file(self):
        path_install_path = self.get_install_path()
        path_package = path_install_path / 'package.json'

        with path_package.open(mode='rb') as file_object:
            return json.load(file_object)

    def _install(self, include_dependencies=False):
        self.get_metadata()
        print(_('Downloading... '), end='')
        sys.stdout.flush()
        self.download()
        print(_('Verifying... '), end='')
        sys.stdout.flush()
        self.verify()
        print(_('Extracting... '), end='')
        sys.stdout.flush()
        self.extract()

        if include_dependencies:
            for name, version_string in self.version_metadata.get('dependencies', {}).items():
                dependency = JavaScriptDependency(
                    name=name, version_string=version_string
                )
                dependency.install(include_dependencies=False)

    def extract(self, replace_list=None):
        with TemporaryDirectory() as temporary_directory:
            path_compressed_file = self.get_tar_file_path()

            with tarfile.open(name=force_text(s=path_compressed_file), mode='r') as file_object:
                file_object.extractall(path=temporary_directory)

            self.patch_files(path=temporary_directory, replace_list=replace_list)

            path_install = self.get_install_path()

            # Clear the installation path of previous content
            shutil.rmtree(path=force_text(s=path_install), ignore_errors=True)

            # Scoped packages are nested under a parent directory
            # create it to avoid rename errors.
            path_install.mkdir(parents=True)

            # Copy the content under the dependency's extracted content folder
            # 'package' to the final location.
            # We do a copy and delete instead of move because os.rename doesn't
            # support renames across filesystems.
            path_uncompressed_package = Path(temporary_directory, 'package')
            shutil.rmtree(path=force_text(s=path_install))
            shutil.copytree(
                src=force_text(s=path_uncompressed_package),
                dst=force_text(s=path_install)
            )

            # Clean up temporary directory used for download
            shutil.rmtree(path=self.path_cache, ignore_errors=True)

    def download(self):
        self.path_cache = mkdtemp()

        with requests.get(self.version_metadata['dist']['tarball'], stream=True) as response:
            response.raise_for_status()
            with self.get_tar_file_path().open(mode='wb') as file_object:
                shutil.copyfileobj(fsrc=response.raw, fdst=file_object)

    def get_best_version(self):
        versions = self.versions
        version_string = self.version_string

        return max_satisfying(
            versions=versions, range_=version_string, loose=True
        )

    def get_copyright(self):
        path_install_path = self.get_install_path()

        for entry in path_install_path.glob(pattern='LICENSE*'):
            with entry.open(mode='rb') as file_object:
                return force_text(s=file_object.read())

        copyright_text = []

        try:
            package_info = self._read_package_file()
        except FileNotFoundError:
            return super().get_copyright()
        else:
            copyright_text.append(
                package_info.get('license') or package_info.get(
                    'licenses'
                )[0]['type']
            )
            author = package_info.get('author', {})

            try:
                author = author.get('name')
            except AttributeError:
                pass

            copyright_text.append(author or '')
            return '\n'.join(copyright_text)

    def get_help_text(self):
        description = None

        try:
            description = self._read_package_file().get('description')
        except FileNotFoundError:
            return super().get_help_text()
        else:
            return description

    def get_install_path(self):
        app = apps.get_app_config(app_label=self.app_label)
        result = Path(
            app.path, 'static', self.static_folder or app.label,
            'node_modules', self.name
        )
        return result

    def get_metadata(self):
        response = requests.get(url=self.get_url())
        self.package_metadata = response.json()
        self.versions = self.package_metadata['versions'].keys()
        self.version_best = self.get_best_version()
        try:
            self.version_metadata = self.package_metadata['versions'][
                self.version_best
            ]
        except KeyError:
            raise DependenciesException(
                'Best version for dependency %s is not found in '
                'upstream repository.', self.version_best
            )

    def get_tar_file_path(self):
        return Path(
            self.path_cache, self.get_tar_filename()
        )

    def get_tar_filename(self):
        return furl(
            self.version_metadata['dist']['tarball']
        ).path.segments[-1]

    def get_url(self):
        url = furl(self.repository.url)
        url.path.segments = url.path.segments + [self.name]
        return url.tostr()

    def verify(self):
        path_tar_file = self.get_tar_file_path()

        try:
            integrity = self.version_metadata['dist']['integrity']
        except KeyError:
            algorithm_name = 'sha1'
            integrity_value = self.version_metadata['dist']['shasum']
        else:
            algorithm_name, integrity_value = integrity.split('-', 1)

        try:
            algorithm_class = HashAlgorithm.get(name=algorithm_name)
        except KeyError:
            raise DependenciesException(
                'Unknown hash algorithm: %s', algorithm_name
            )

        with path_tar_file.open(mode='rb') as file_object:
            algorithm_object = algorithm_class(file_object=file_object)
            algorithm_object.calculate()

        if algorithm_object.get_digest() != integrity_value:
            path_tar_file.unlink()
            raise DependenciesException(
                'Hash of downloaded dependency package "%s" doesn\'t match '
                'online version.', self.get_label_full()
            )


class PythonVersion:
    def __init__(self, string):
        self.version = version.parse(string)

    def __lt__(self, other):
        return self.version < other.version


class PythonDependency(Dependency):
    class_name = 'python'
    class_name_help_text = _(
        'Python packages downloaded from PyPI.'
    )
    class_name_verbose_name = _('Python')
    provider_class = PyPIRespository

    def __init__(self, *args, **kwargs):
        self.copyright_attribute = kwargs.pop('copyright_attribute', None)
        super().__init__(*args, **kwargs)

    def _check(self):
        try:
            return pkg_resources.get_distribution(
                dist='{}{}'.format(self.name, self.version_string)
            ) is not None
        except pkg_resources.DistributionNotFound:
            return False
        except pkg_resources.VersionConflict:
            return False

    def get_copyright(self):
        if self.copyright_attribute:
            return import_string(dotted_path=self.copyright_attribute)
        else:
            return super().get_copyright()

    def get_latest_version(self):
        url = 'https://pypi.python.org/pypi/{}/json'.format(self.name)
        response = requests.get(url=url)
        versions = list(response.json()['releases'])
        versions.sort(key=PythonVersion)
        return versions[-1]

    def is_latest_version(self):
        return self.version_string == '=={}'.format(self.get_latest_version())


class GoogleFontDependency(Dependency):
    class_name = 'google_font'
    class_name_help_text = _(
        'Fonts downloaded from fonts.googleapis.com.'
    )
    class_name_verbose_name = _('Google font')
    provider_class = GoogleFontsProvider
    user_agents = {
        'woff2': 'Mozilla/5.0 (Windows NT 6.1; rv:60.0) Gecko/20100101 Firefox/60.0',
        'woff': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.101 Safari/537.36',
        'ttf': 'Mozilla/5.0 (Linux; U; Android 2.2; en-us; DROID2 GLOBAL Build/S273) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1',
    }

    def __init__(self, *args, **kwargs):
        self.url = kwargs.pop('url')
        self.static_folder = kwargs.pop('static_folder', None)
        super().__init__(*args, **kwargs)

    def _check(self):
        return self.get_install_path().exists()

    def _install(self):
        print(_('Downloading... '), end='')
        sys.stdout.flush()
        self.download()
        print(_('Extracting... '), end='')
        sys.stdout.flush()
        self.extract()

    def download(self):
        self.path_cache = Path(mkdtemp())
        # Use .css to keep the same ContentType, otherwise the webserver
        # will use the generic octet and the browser will ignore the import
        # https://www.w3.org/TR/2013/CR-css-cascade-3-20131003/#content-type
        self.path_import_file = self.path_cache / 'import.css'

        self.font_files = []

        with self.path_import_file.open(mode='w') as file_object:
            for agent_name, agent_string in self.user_agents.items():
                import_file = force_text(
                    s=requests.get(
                        self.url, headers={
                            'User-Agent': agent_string
                        }
                    ).content
                )

                for line in import_file.split('\n'):
                    if 'url' in line:
                        font_url = line.split(' ')[-2][4:-1]
                        url = furl(force_text(s=font_url))
                        font_filename = url.path.segments[-1]

                        path_font_filename = self.path_cache / font_filename
                        with path_font_filename.open(mode='wb') as font_file_object:
                            with requests.get(font_url, stream=True) as response:
                                # Use response.content instead of response.raw
                                # to allow requests to handle gzip and deflate
                                # content.
                                # https://2.python-requests.org/en/master/user/quickstart/#binary-response-content
                                shutil.copyfileobj(
                                    fsrc=BytesIO(response.content),
                                    fdst=font_file_object
                                )

                        line = line.replace(font_url, font_filename)

                    file_object.write(line)

    def extract(self, replace_list=None):
        path_install = self.get_install_path()

        # Clear the installation path of previous content
        shutil.rmtree(path=force_text(s=path_install), ignore_errors=True)

        shutil.copytree(
            src=force_text(s=self.path_cache), dst=force_text(s=path_install)
        )
        shutil.rmtree(path=force_text(s=self.path_cache), ignore_errors=True)

    def get_install_path(self):
        app = apps.get_app_config(app_label=self.app_label)
        result = Path(
            app.path, 'static', self.static_folder or app.label,
            'google_fonts', self.name
        )
        return result


DependencyGroup(
    attribute_name='app_label', label=_('Declared in app'), help_text=_(
        'Show dependencies by the app that declared them.'
    ), name='app'
)
DependencyGroup(
    attribute_name='class_name', label=_('Class'), help_text=_(
        'Show the different classes of dependencies. Classes are usually '
        'divided by language or the file types of the dependency.'
    ), name='class'
)
DependencyGroup(
    attribute_name='check_string', label=_('State'), help_text=_(
        'Show the different states of the dependencies. True means that the '
        'dependencies is installed and is of a correct version. False means '
        'the dependencies is missing or an incorrect version is present.'
    ), name='state'
)
DependencyGroup(
    allow_multiple=True, attribute_name='get_environments',
    label=_('Environments'), help_text=_(
        'Dependencies required for an environment might not be required for '
        'another. Example environments: Production, Development.'
    ), name='environment'
)



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..permissions import permission_template_sandbox

from .mixins import DocumentTemplateSandboxViewTestMixin


class DocumentTemplateSandboxViewTestCase(
    DocumentTemplateSandboxViewTestMixin, GenericDocumentViewTestCase
):
    def test_document_template_sanbox_get_view_no_permission(self):
        response = self._request_document_template_sandbox_get_view()
        self.assertEqual(response.status_code, 404)

    def test_document_template_sanbox_get_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_template_sandbox
        )

        response = self._request_document_template_sandbox_get_view()
        self.assertEqual(response.status_code, 200)

    def test_document_template_sanbox_post_view_no_permission(self):
        response = self._request_document_template_sandbox_post_view()
        self.assertEqual(response.status_code, 404)

    def test_document_template_sanbox_post_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_template_sandbox
        )

        response = self._request_document_template_sandbox_post_view()
        self.assertEqual(response.status_code, 302)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('tags', '0005_auto_20150718_0616'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentTag',
            fields=[
            ],
            options={
                'verbose_name': 'Document tag',
                'proxy': True,
                'verbose_name_plural': 'Document tags',
            },
            bases=('tags.tag',),
        ),
    ]



from mayan.apps.documents.serializers.document_serializers import DocumentSerializer
from mayan.apps.rest_api import serializers

from .models import DuplicateTargetDocument


class DuplicateTargetDocumentSerializer(DocumentSerializer):
    backend = serializers.CharField(read_only=True)

    class Meta(DocumentSerializer.Meta):
        fields = sorted(
            DocumentSerializer.Meta.fields + (
                'backend',
            )
        )
        model = DuplicateTargetDocument



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_view
)
from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_type_smart_links, icon_smart_link_condition_create,
    icon_smart_link_condition_delete, icon_smart_link_condition_edit,
    icon_smart_link_condition_list, icon_smart_link_create,
    icon_smart_link_delete, icon_smart_link_document_types,
    icon_smart_link_edit, icon_smart_link_instances_for_document,
    icon_smart_link_list, icon_smart_link_setup
)
from .permissions import (
    permission_smart_link_create, permission_smart_link_delete,
    permission_smart_link_edit, permission_smart_link_view
)


link_document_type_smart_links = Link(
    args='resolved_object.pk', icon=icon_document_type_smart_links,
    permissions=(permission_document_type_edit,), text=_('Smart links'),
    view='linking:document_type_smart_links'
)
link_smart_link_condition_create = Link(
    args='object.pk', icon=icon_smart_link_condition_create,
    permissions=(permission_smart_link_edit,), text=_('Create condition'),
    view='linking:smart_link_condition_create'
)
link_smart_link_condition_delete = Link(
    args='resolved_object.pk', icon=icon_smart_link_condition_delete,
    permissions=(permission_smart_link_edit,), tags='dangerous',
    text=_('Delete'), view='linking:smart_link_condition_delete'
)
link_smart_link_condition_edit = Link(
    args='resolved_object.pk', icon=icon_smart_link_condition_edit,
    permissions=(permission_smart_link_edit,), text=_('Edit'),
    view='linking:smart_link_condition_edit'
)
link_smart_link_condition_list = Link(
    args='object.pk', icon=icon_smart_link_condition_list,
    permissions=(permission_smart_link_edit,), text=_('Conditions'),
    view='linking:smart_link_condition_list'
)
link_smart_link_create = Link(
    icon=icon_smart_link_create,
    permissions=(permission_smart_link_create,),
    text=_('Create new smart link'), view='linking:smart_link_create'
)
link_smart_link_delete = Link(
    args='object.pk', icon=icon_smart_link_delete,
    permissions=(permission_smart_link_delete,), tags='dangerous',
    text=_('Delete'), view='linking:smart_link_delete'
)
link_smart_link_document_types = Link(
    args='object.pk', icon=icon_smart_link_document_types,
    permissions=(permission_smart_link_edit,), text=_('Document types'),
    view='linking:smart_link_document_types'
)
link_smart_link_edit = Link(
    args='object.pk', icon=icon_smart_link_edit,
    permissions=(permission_smart_link_edit,), text=_('Edit'),
    view='linking:smart_link_edit'
)
link_smart_link_instance_view = Link(
    args=('document.pk', 'object.pk',),
    permissions=(permission_smart_link_view,), text=_('Documents'),
    view='linking:smart_link_instance_view'
)
link_smart_link_instances_for_document = Link(
    args='resolved_object.pk',
    icon=icon_smart_link_instances_for_document,
    permissions=(permission_document_view,), text=_('Smart links'),
    view='linking:smart_link_instances_for_document'
)
link_smart_link_list = Link(
    icon=icon_smart_link_list,
    permissions=(permission_smart_link_create,), text=_('Smart links'),
    view='linking:smart_link_list'
)
link_smart_link_setup = Link(
    icon=icon_smart_link_setup,
    permissions=(permission_smart_link_create,), text=_('Smart links'),
    view='linking:smart_link_list'
)



from ..models import DocumentVersionPageOCRContent

from .literals import (
    TEST_DOCUMENT_VERSION_OCR_CONTENT,
    TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
)


class DocumentTypeOCRSettingsAPIViewTestMixin:
    def _request_test_document_type_ocr_settings_details_api_view(self):
        return self.get(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self.test_document_type.pk}
        )

    def _request_test_document_type_ocr_settings_patch_api_view(self):
        return self.patch(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self.test_document_type.pk},
            data={'auto_ocr': True}
        )

    def _request_test_document_type_ocr_settings_put_api_view(self):
        return self.put(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self.test_document_type.pk},
            data={'auto_ocr': True}
        )


class DocumentTypeOCRViewTestMixin:
    def _request_test_document_type_ocr_settings_view(self):
        return self.get(
            viewname='ocr:document_type_ocr_settings', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_document_type_ocr_submit_view(self):
        return self.post(
            viewname='ocr:document_type_submit', data={
                'document_type': [self.test_document_type.pk]
            }
        )


class DocumentOCRAPIViewTestMixin:
    def _request_test_document_ocr_submit_api_view(self):
        return self.post(
            viewname='rest_api:document-ocr-submit-view',
            kwargs={'document_id': self.test_document.pk}
        )


class DocumentVersionOCRAPIViewTestMixin:
    def _request_test_document_version_ocr_submit_api_view(self):
        return self.post(
            viewname='rest_api:document-version-ocr-submit-view', kwargs={
                'document_id': self.test_document.pk,
                'document_version_id': self.test_document_version.pk
            }
        )


class DocumentVersionPageOCRAPIViewTestMixin:
    def _request_test_document_version_page_ocr_content_detail_api_view_via_get(self):
        return self.get(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self.test_document.pk,
                'document_version_id': self.test_document_version.pk,
                'document_version_page_id': self.test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_ocr_content_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self.test_document.pk,
                'document_version_id': self.test_document_version.pk,
                'document_version_page_id': self.test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )

    def _request_test_document_version_page_ocr_content_edit_api_view_via_put(self):
        return self.put(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self.test_document.pk,
                'document_version_id': self.test_document_version.pk,
                'document_version_page_id': self.test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )


class DocumentVersionOCRTestMixin:
    auto_create_test_document_version_ocr_content = False

    def setUp(self):
        super().setUp()

        if self.auto_create_test_document_version_ocr_content:
            self._create_test_document_version_ocr_content()

    def _create_test_document_version_ocr_content(self):
        DocumentVersionPageOCRContent.objects.create(
            document_version_page=self.test_document_version_page,
            content=TEST_DOCUMENT_VERSION_OCR_CONTENT
        )


class DocumentVersionOCRViewTestMixin:
    def _request_test_document_version_ocr_content_view(self):
        return self.get(
            viewname='ocr:document_version_ocr_content_view', kwargs={
                'document_version_id': self.test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_content_delete_single_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_content_delete_single',
            kwargs={
                'document_version_id': self.test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_content_delete_multiple_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_content_delete_multiple',
            data={
                'id_list': self.test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_error_list_view(self):
        return self.get(
            viewname='ocr:document_version_ocr_error_list', kwargs={
                'document_version_id': self.test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_submit_single_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_submit_single', kwargs={
                'document_version_id': self.test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_submit_multiple_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_submit_multiple', data={
                'id_list': self.test_document_version.pk,
            }
        )

    def _request_test_document_version_ocr_download_view(self):
        return self.get(
            viewname='ocr:document_version_ocr_download', kwargs={
                'document_version_id': self.test_document_version.pk
            }
        )


class DocumentVersionPageOCRViewTestMixin:
    def _request_test_document_version_page_ocr_content_detail_view(self):
        return self.get(
            viewname='ocr:document_version_page_ocr_content_detail_view', kwargs={
                'document_version_page_id': self.test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_ocr_content_edit_view(self):
        return self.post(
            viewname='ocr:document_version_page_ocr_content_edit_view', kwargs={
                'document_version_page_id': self.test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase


from .mixins.workflow_template_mixins import (
    WorkflowTaskTestCaseMixin, WorkflowTemplateTestMixin
)


class WorkflowTaskTestCase(
    WorkflowTaskTestCaseMixin, WorkflowTemplateTestMixin,
    GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()

    def test_task_launch_all_workflows(self):
        workflow_instance_count = self.test_document.workflows.count()

        self._execute_task_launch_all_workflows()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count + 1
        )

    def test_trashed_document_task_launch_all_workflows(self):
        workflow_instance_count = self.test_document.workflows.count()

        self.test_document.delete()

        self._execute_task_launch_all_workflows()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count
        )

    def test_task_launch_workflow(self):
        workflow_instance_count = self.test_document.workflows.count()

        self._execute_task_launch_workflow()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count + 1
        )

    def test_trashed_document_task_launch_workflow(self):
        workflow_instance_count = self.test_document.workflows.count()

        self.test_document.delete()

        self._execute_task_launch_workflow()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count
        )

    def test_task_launch_workflow_for(self):
        workflow_instance_count = self.test_document.workflows.count()

        self._execute_task_launch_workflow_for()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count + 1
        )

    def test_trashed_document_task_launch_workflow_for(self):
        workflow_instance_count = self.test_document.workflows.count()

        self.test_document.delete()

        with self.assertRaises(expected_exception=Document.DoesNotExist):
            self._execute_task_launch_workflow_for()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count
        )

    def test_task_launch_all_workflow_for(self):
        workflow_instance_count = self.test_document.workflows.count()

        self._execute_task_launch_all_workflow_for()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count + 1
        )

    def test_trashed_document_task_launch_all_workflow_for(self):
        workflow_instance_count = self.test_document.workflows.count()

        self.test_document.delete()

        with self.assertRaises(expected_exception=Document.DoesNotExist):
            self._execute_task_launch_all_workflow_for()

        self.assertEqual(
            self.test_document.workflows.count(), workflow_instance_count
        )



from django.conf import settings
from django.db import migrations


def code_user_otp_data_initialize(apps, schema_editor):
    User = apps.get_model(settings.AUTH_USER_MODEL)
    UserOTPData = apps.get_model(
        app_label='authentication_otp', model_name='UserOTPData'
    )

    for user in User.objects.using(alias=schema_editor.connection.alias).all():
        UserOTPData.objects.using(
            alias=schema_editor.connection.alias
        ).get_or_create(user=user)


class Migration(migrations.Migration):
    dependencies = [
        ('authentication_otp', '0001_initial')
    ]

    operations = [
        migrations.RunPython(code=code_user_otp_data_initialize)
    ]



from mayan.apps.smart_settings.classes import Setting, SettingNamespaceMigration
from mayan.apps.smart_settings.utils import smart_yaml_load


class SourcesSettingMigration(SettingNamespaceMigration):
    """
    0001 to 0002: Backend arguments are no longer quoted but YAML valid too.
                  Changed in version 3.3.
    0002 to 0003: New settings for source cache storage.
                  SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND,
                  SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS are
                  now SOURCES_CACHE_STORAGE_BACKEND and
                  SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS

    """
    def sources_staging_file_cache_storage_backend_arguments_0001(self, value):
        return smart_yaml_load(value=value)

    def sources_cache_storage_backend_0002(self, value):
        # Get the setting by its new global name.
        setting = Setting.get(
            global_name='SOURCES_CACHE_STORAGE_BACKEND'
        )
        # Load the value from the setting's old global name.
        setting.cache_value(
            global_name='SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND'
        )
        return setting.value

    def sources_cache_storage_backend_arguments_0002(self, value):
        # Get the setting by its new global name.
        setting = Setting.get(
            global_name='SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        # Load the value from the setting's old global name.
        setting.cache_value(
            global_name='SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        return setting.value



from mayan.apps.rest_api import serializers


class SearchFieldSerializer(serializers.Serializer):
    field = serializers.CharField(read_only=True)
    label = serializers.CharField(read_only=True)


class SearchModelSerializer(serializers.Serializer):
    app_label = serializers.CharField(read_only=True)
    model_name = serializers.CharField(read_only=True)
    pk = serializers.CharField(read_only=True)
    search_fields = SearchFieldSerializer(many=True, read_only=True)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('checkouts', '0004_auto_20150617_0330'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentcheckout',
            name='document',
            field=models.OneToOneField(
                on_delete=models.CASCADE, to='documents.Document',
                verbose_name='Document'
            ),
        ),
    ]



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase
from mayan.apps.user_management.tests.mixins import GroupTestMixin
from mayan.apps.user_management.permissions import (
    permission_group_edit, permission_group_view
)

from ..classes import Permission
from ..events import event_role_created, event_role_edited
from ..models import Role
from ..permissions import (
    permission_role_create, permission_role_delete,
    permission_role_edit, permission_role_view
)

from .mixins import (
    PermissionAPIViewTestMixin, RoleAPIViewTestMixin,
    RoleGroupAPIViewTestMixin, RolePermissionAPIViewTestMixin, RoleTestMixin
)


class PermissionAPIViewTestCase(PermissionAPIViewTestMixin, BaseAPITestCase):
    def setUp(self):
        super().setUp()
        Permission.invalidate_cache()

    def test_permissions_list_api_view(self):
        self._clear_events()

        response = self._request_permissions_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class RoleAPIViewTestCase(
    GroupTestMixin, RoleAPIViewTestMixin, RoleTestMixin, BaseAPITestCase
):
    def test_role_create_api_view_no_permission(self):
        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Role.objects.count(), role_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_role_create)

        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Role.objects.count(), role_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_created.id)

    def test_role_delete_api_view_no_permission(self):
        self._create_test_role()

        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Role.objects.count(), role_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_delete_api_view_with_access(self):
        self._create_test_role()

        self.grant_access(obj=self.test_role, permission=permission_role_delete)

        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Role.objects.count(), role_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_edit_api_view_via_patch_no_permission(self):
        self._create_test_role()

        role_label = self.test_role.label

        self._clear_events()

        response = self._request_test_role_edit_api_view(request_type='patch')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_role.refresh_from_db()
        self.assertEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_edit_api_view_via_patch_with_access(self):
        self._create_test_role()
        self.grant_access(obj=self.test_role, permission=permission_role_edit)

        role_label = self.test_role.label

        self._clear_events()

        response = self._request_test_role_edit_api_view(request_type='patch')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_role.refresh_from_db()
        self.assertNotEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_edit_api_view_via_put_no_permission(self):
        self._create_test_role()

        self._clear_events()

        response = self._request_test_role_edit_api_view(request_type='put')

        role_label = self.test_role.label

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_role.refresh_from_db()
        self.assertEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_edit_api_view_via_put_with_access(self):
        self._create_test_role()
        self.grant_access(obj=self.test_role, permission=permission_role_edit)

        role_label = self.test_role.label

        self._clear_events()

        response = self._request_test_role_edit_api_view(request_type='put')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_role.refresh_from_db()
        self.assertNotEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_roles_list_api_view_no_permission(self):
        self._create_test_role()

        self._clear_events()

        response = self._request_test_role_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_roles_list_api_view_with_access(self):
        self._create_test_role()
        self.grant_access(
            obj=self.test_role, permission=permission_role_view
        )

        self._clear_events()

        response = self._request_test_role_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['label'], self.test_role.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class RoleGroupAPIViewTestCase(
    GroupTestMixin, RoleTestMixin, RoleGroupAPIViewTestMixin,
    BaseAPITestCase
):
    auto_create_role_test_object = True

    def setUp(self):
        super().setUp()
        self._create_test_role()
        self._create_test_group()

    def test_role_group_add_api_view_no_group(self):
        self._clear_events()

        response = self._request_test_role_group_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(self.test_group in self.test_role.groups.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_api_view_with_group_access(self):
        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(self.test_group in self.test_role.groups.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_api_view_with_role_access(self):
        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertFalse(self.test_group in self.test_role.groups.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(self.test_group in self.test_role.groups.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_group_list_api_view_no_group(self):
        self.test_role.groups.add(self.test_group)

        self._clear_events()

        response = self._request_test_role_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_list_api_view_with_group_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group, permission=permission_group_view
        )

        self._clear_events()

        response = self._request_test_role_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_list_api_view_with_role_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role, permission=permission_role_view
        )

        self._clear_events()

        response = self._request_test_role_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_list_api_view_with_full_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group, permission=permission_group_view
        )
        self.grant_access(
            obj=self.test_role, permission=permission_role_view
        )

        self._clear_events()

        response = self._request_test_role_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self.test_group.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_api_view_no_group(self):
        self.test_role.groups.add(self.test_group)

        role_group_count = self.test_role.groups.count()

        self._clear_events()

        response = self._request_test_role_group_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_role.groups.count(), role_group_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_api_view_with_group_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )

        role_group_count = self.test_role.groups.count()

        self._clear_events()

        response = self._request_test_role_group_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_role.groups.count(), role_group_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_api_view_with_role_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        role_group_count = self.test_role.groups.count()

        self._clear_events()

        response = self._request_test_role_group_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self.test_role.groups.count(), role_group_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_api_view_with_full_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        role_group_count = self.test_role.groups.count()

        self._clear_events()

        response = self._request_test_role_group_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self.test_role.groups.count(), role_group_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)


class RolePermissionAPIViewTestCase(
    RoleTestMixin, RolePermissionAPIViewTestMixin, BaseAPITestCase
):
    auto_create_role_test_object = True

    def setUp(self):
        super().setUp()
        self._create_test_role()
        self._create_test_permission()

    def test_role_permission_add_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_role_permission_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(
            self.test_permission.stored_permission in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_add_api_view_with_access(self):
        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_permission_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self.test_permission.stored_permission in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_permission.stored_permission
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_permission_list_api_view_no_permission(self):
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self._clear_events()

        response = self._request_test_role_permission_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_list_api_view_with_access(self):
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self.grant_access(
            obj=self.test_role, permission=permission_role_view
        )

        self._clear_events()

        response = self._request_test_role_permission_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['pk'],
            self.test_permission.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_remove_api_view_no_permission(self):
        self.test_role.permissions.add(
            self.test_permission.stored_permission
        )

        role_permission_count = self.test_role.permissions.count()

        self._clear_events()

        response = self._request_test_role_permission_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_role.permissions.count(), role_permission_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_remove_api_view_with_access(self):
        self.test_role.permissions.add(
            self.test_permission.stored_permission
        )

        self.grant_access(
            obj=self.test_role, permission=permission_role_edit
        )

        role_permission_count = self.test_role.permissions.count()

        self._clear_events()

        response = self._request_test_role_permission_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self.test_role.permissions.count(), role_permission_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_permission.stored_permission
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)



import logging

from django.db import models, transaction
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext, ugettext_lazy as _

from mptt.fields import TreeForeignKey
from mptt.models import MPTTModel

from mayan.apps.acls.models import AccessControlList
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models import Document, DocumentType
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.templating.classes import Template

from .events import event_index_template_created, event_index_template_edited
from .managers import (
    DocumentIndexInstanceNodeManager, IndexInstanceManager,
    IndexTemplateManager
)

logger = logging.getLogger(name=__name__)


class IndexTemplate(ExtraDataModelMixin, models.Model):
    """
    Parent model that defines an index and hold all the relationship for its
    template and instance when resolved.
    """
    label = models.CharField(
        help_text=_('Short description of this index.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    slug = models.SlugField(
        help_text=_(
            'This value will be used by other apps to reference this index.'
        ), max_length=128, unique=True, verbose_name=_('Slug')
    )
    enabled = models.BooleanField(
        default=True,
        help_text=_(
            'Causes this index to be visible and updated when document data '
            'changes.'
        ),
        verbose_name=_('Enabled')
    )
    document_types = models.ManyToManyField(
        related_name='index_templates', to=DocumentType,
        verbose_name=_('Document types')
    )

    objects = IndexTemplateManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Index template')
        verbose_name_plural = _('Index templates')

    def __str__(self):
        return self.label

    def document_types_add(self, queryset, _event_actor=None):
        for document_type in queryset:
            self.document_types.add(document_type)

            event_index_template_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor, target=self
            )

    def document_types_remove(self, queryset, _event_actor=None):
        for document_type in queryset:
            self.document_types.remove(document_type)

            event_index_template_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor, target=self
            )

    def delete_index_instance_nodes(self):
        try:
            IndexInstanceNode.objects.filter(index_template_node__index=self).delete()
        except IndexInstanceNode.DoesNotExist:
            """Empty index, ignore this exception."""

    def get_absolute_url(self):
        return reverse(
            viewname='indexing:index_template_view', kwargs={
                'index_template_id': self.pk
            }
        )

    def get_document_types_names(self):
        return ', '.join(
            [
                force_text(s=document_type) for document_type in self.document_types.all()
            ] or ['None']
        )

    @property
    def index_template_root_node(self):
        """
        Return the root node for this index.
        """
        return self.index_template_nodes.get(parent=None)

    def natural_key(self):
        return (self.slug,)

    def rebuild(self):
        """
        Delete and reconstruct the index by deleting of all its instance nodes
        and recreating them for the documents whose types are associated with
        this index
        """
        if self.enabled:
            self.delete_index_instance_nodes()

            # Create the new root index instance node.
            self.index_template_root_node.index_instance_nodes.create()

            index_instance = IndexInstance.objects.get(pk=self.pk)
            index_instance.index_instance_root_node
            # Re-index each document with a type associated with this index.
            for document in Document.objects.filter(document_type__in=self.document_types.all()):
                # Evaluate each index template node for each document
                # associated with this index.
                index_instance.document_add(document=document)

    def reset(self):
        self.delete_index_instance_nodes()

        # Create the new root index instance node.
        self.index_template_root_node.initialize_index_instance_root_node()

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_index_template_created,
            'target': 'self',
        },
        edited={
            'event': event_index_template_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        is_new = not self.pk
        super().save(*args, **kwargs)
        if is_new:
            # Automatically create the root index template node.
            IndexTemplateNode.objects.get_or_create(parent=None, index=self)

        self.index_template_root_node.initialize_index_instance_root_node()


class IndexTemplateNode(MPTTModel):
    """
    The template to generate an index. Each entry represents a level in a
    hierarchy of levels. Each level can contain further levels or a list of
    documents but not both.
    """
    parent = TreeForeignKey(
        blank=True, help_text=_('Parent index template node of this node.'),
        null=True, on_delete=models.CASCADE,
        related_name='children', to='self'
    )
    index = models.ForeignKey(
        on_delete=models.CASCADE, related_name='index_template_nodes',
        to=IndexTemplate, verbose_name=_('Index')
    )
    expression = models.TextField(
        help_text=_(
            'Enter a template to render. Use Django\'s default templating '
            'language.'
        ),
        verbose_name=_('Indexing expression')
    )
    enabled = models.BooleanField(
        default=True,
        help_text=_(
            'Causes this node to be visible and updated when document data '
            'changes.'
        ),
        verbose_name=_('Enabled')
    )
    link_documents = models.BooleanField(
        default=False,
        help_text=_(
            'Check this option to have this node act as a container for '
            'documents and not as a parent for further nodes.'
        ),
        verbose_name=_('Link documents')
    )

    class Meta:
        verbose_name = _('Index template node')
        verbose_name_plural = _('Index template nodes')

    def __str__(self):
        if self.is_root_node():
            return ugettext('Root')
        else:
            return self.expression

    def get_index_instance_root_node(self):
        return self.index_instance_nodes.get(parent=None)

    def initialize_index_instance_root_node(self):
        self.index_instance_nodes.get_or_create(parent=None)


class IndexInstance(IndexTemplate):
    """
    Model that represents an evaluated index. This is an index whose nodes
    have been evaluated against a series of documents. If is a proxy model
    at the moment.
    """
    objects = IndexInstanceManager()

    class Meta:
        proxy = True
        verbose_name = _('Index instance')
        verbose_name_plural = _('Index instances')

    def delete_empty_nodes(self, acquire_lock=True):
        if acquire_lock:
            try:
                if acquire_lock:
                    lock_index_instance = LockingBackend.get_backend().acquire_lock(
                        name=self.get_lock_string()
                    )
            except LockError:
                raise
            else:
                try:
                    return self._delete_empty_nodes()
                finally:
                    lock_index_instance.release()
        else:
            return self._delete_empty_nodes()

    def _delete_empty_nodes(self):
        with transaction.atomic():
            while True:
                queryset = IndexInstanceNode.objects.filter(children=None, index_template_node__link_documents=True, documents=None)
                if queryset.exists():
                    queryset.delete()
                else:
                    break

    def document_add(self, document):
        """
        Method to start the indexing process for a document. The entire process
        happens inside one transaction. The document is first removed from all
        the index nodes to which it already belongs. The different index
        templates that match this document's type are evaluated and for each
        result a node is fetched or created and the document is added to that
        node.
        """
        logger.debug('Index; Indexing document: %s', document)

        if Document.valid.filter(pk=document.pk) and self.enabled and self.document_types.filter(pk=document.document_type.pk).exists():
            try:
                locking_backend = LockingBackend.get_backend()

                lock_index_instance = locking_backend.acquire_lock(
                    name=self.get_lock_string()
                )
            except LockError:
                raise
            else:
                try:
                    lock_document = locking_backend.acquire_lock(
                        name=self.get_document_lock_string(document=document)
                    )
                except LockError:
                    raise
                else:
                    try:
                        self.initialize_index_instance_root_node_node()

                        self.document_remove(acquire_lock=False, document=document)

                        index_instance_node_parent = self.index_instance_root_node

                        self._document_add(document=document, index_instance_node_parent=index_instance_node_parent)
                    finally:
                        lock_document.release()
                finally:
                    lock_index_instance.release()

    def _document_add(self, document, index_instance_node_parent):
        for index_template_node in index_instance_node_parent.index_template_node.get_children().filter(enabled=True):
            try:
                template = Template(
                    template_string=index_template_node.expression
                )
                result = template.render(
                    context={'document': document}
                )
            except Exception as exception:
                logger.debug('Evaluating error: %s', exception)
                error_message = _(
                    'Error indexing document: %(document)s; expression: '
                    '%(expression)s; %(exception)s'
                ) % {
                    'document': document,
                    'expression': index_template_node.expression,
                    'exception': exception
                }
                logger.debug(error_message)
            else:
                logger.debug('Evaluation result: %s', result)

                if result:
                    index_instance_node, created = index_template_node.index_instance_nodes.get_or_create(
                        parent=index_instance_node_parent,
                        value=result
                    )

                    if index_template_node.link_documents:
                        index_instance_node.documents.add(document)

                    self._document_add(document=document, index_instance_node_parent=index_instance_node)

    def document_remove(self, document, acquire_lock=True):
        if Document.valid.filter(pk=document.pk) and self.enabled and self.document_types.filter(pk=document.document_type.pk).exists():
            if acquire_lock:
                try:
                    lock_index_instance = LockingBackend.get_backend().acquire_lock(
                        name=self.get_lock_string()
                    )
                except LockError:
                    raise
                else:
                    try:
                        lock_document = LockingBackend.get_backend().acquire_lock(
                            name=self.get_document_lock_string(document=document)
                        )
                    except LockError:
                        raise
                    else:
                        try:
                            return self._document_remove(document=document)
                        finally:
                            lock_document.release()
                    finally:
                        lock_index_instance.release()
            else:
                return self._document_remove(document=document)

    def _document_remove(self, document):
        with transaction.atomic():
            document.index_instance_nodes.through._meta.model.objects.filter(document=document, indexinstancenode__index_template_node__index=self).delete()
            self.delete_empty_nodes(acquire_lock=False)

    def get_absolute_url(self):
        try:
            index_instance_root_node = self.index_instance_root_node
        except IndexInstanceNode.DoesNotExist:
            return '#'
        else:
            return reverse(
                viewname='indexing:index_instance_node_view', kwargs={
                    'index_instance_node_id': index_instance_root_node.pk
                }
            )

    def get_children(self):
        return self.index_instance_root_node.get_children()

    def get_document_lock_string(self, document):
        return 'indexing:document_{}'.format(document.pk)

    def get_descendants(self):
        return self.index_instance_root_node.get_descendants()

    def get_descendants_count(self):
        return self.index_instance_root_node.get_descendants_count()

    get_descendants_count.help_text = _(
        'Total number of nodes with unique values this item contains.'
    )

    def get_descendants_document_count(self, user):
        return self.index_instance_root_node.get_descendants_document_count(
            user=user
        )

    get_descendants_document_count.help_text = _(
        'Total number of unique documents this item contains.'
    )

    def get_lock_string(self):
        return 'indexing:index_instance_{}'.format(self.pk)

    def get_level_count(self):
        return self.index_instance_root_node.get_level_count()

    get_level_count.help_text = _(
        'Total number of node levels this item contains.'
    )

    @property
    def index_instance_root_node(self):
        return self.index_template_root_node.get_index_instance_root_node()

    def initialize_index_instance_root_node_node(self):
        return self.index_template_root_node.initialize_index_instance_root_node()


class IndexInstanceNode(MPTTModel):
    """
    This model represent one instance node from a index template node. That is
    a node template that has been evaluated against a document and the result
    from that evaluation is this node's stored values. Instances of this
    model also point to the original node template.
    """
    parent = TreeForeignKey(
        blank=True, null=True, on_delete=models.CASCADE,
        related_name='children', to='self'
    )
    index_template_node = models.ForeignKey(
        on_delete=models.CASCADE, related_name='index_instance_nodes',
        to=IndexTemplateNode, verbose_name=_('Index template node')
    )
    value = models.CharField(
        blank=True, db_index=True, max_length=128, verbose_name=_('Value')
    )
    documents = models.ManyToManyField(
        related_name='index_instance_nodes', to=Document,
        verbose_name=_('Documents')
    )

    class Meta:
        unique_together = ('index_template_node', 'parent', 'value')
        verbose_name = _('Index instance node')
        verbose_name_plural = _('Indexes instances node')

    def __str__(self):
        return self.value

    def get_absolute_url(self):
        return reverse(
            viewname='indexing:index_instance_node_view', kwargs={
                'index_instance_node_id': self.pk
            }
        )

    def get_children_count(self):
        return self.get_children().count()

    def get_descendants_count(self):
        return self.get_descendants().count()

    get_descendants_count.help_text = IndexInstance.get_descendants_count.help_text

    def get_descendants_document_count(self, user):
        queryset = Document.valid.filter(
            index_instance_nodes__in=self.get_descendants(
                include_self=True
            )
        ).distinct()

        return AccessControlList.objects.restrict_queryset(
            permission=permission_document_view,
            queryset=queryset, user=user
        ).count()

    get_descendants_document_count.help_text = IndexInstance.get_descendants_document_count.help_text

    def get_documents(self):
        return Document.valid.filter(pk__in=self.documents.values('pk'))

    def get_full_path(self):
        result = []
        for node in self.get_ancestors(include_self=True):
            if node.is_root_node():
                result.append(force_text(s=self.index()))
            else:
                result.append(force_text(s=node))

        return ' / '.join(result)
    get_full_path.help_text = _(
        'The path to the index including all ancestors.'
    )
    get_full_path.short_description = _('Full path')

    def get_level_count(self):
        return self.get_descendants().values('level').distinct().count()

    get_level_count.help_text = IndexInstance.get_level_count.help_text

    def index(self):
        """
        Return's the index instance of this node instance.
        """
        return IndexInstance.objects.get(pk=self.index_template_node.index.pk)


class DocumentIndexInstanceNode(IndexInstanceNode):
    """
    Proxy model of node instance. It is used to represent the node instance
    in which a document is currently located. It is used to aid in column
    registration. The inherited methods of this model should not be used.
    """
    objects = DocumentIndexInstanceNodeManager()

    class Meta:
        proxy = True
        verbose_name = _('Document index node instance')
        verbose_name_plural = _('Document indexes node instances')


class IndexInstanceNodeSearchResult(IndexInstanceNode):
    class Meta:
        proxy = True
        verbose_name = _('Index instance node')
        verbose_name_plural = _('Index instance nodes')



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0070_auto_20201128_0249'),
    ]

    operations = [
        migrations.RenameField(
            model_name='document',
            old_name='date_added',
            new_name='datetime_created',
        ),
    ]



from django.db.models import Q

from mayan.apps.converter.layers import layer_saved_transformations

from ...literals import PAGE_RANGE_ALL
from ...models.document_file_models import DocumentFile

from ..literals import (
    TEST_DOCUMENT_FILE_ACTION, TEST_DOCUMENT_FILE_COMMENT,
    TEST_DOCUMENT_FILE_COMMENT_EDITED, TEST_DOCUMENT_FILE_FILENAME_EDITED,
    TEST_SMALL_DOCUMENT_PATH, TEST_TRANSFORMATION_ARGUMENT,
    TEST_TRANSFORMATION_CLASS
)


class DocumentFileAPIViewTestMixin:
    def _request_test_document_file_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documentfile-detail', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document.file_latest.pk
            }
        )

    def _request_test_document_file_detail_api_view(self):
        return self.get(
            viewname='rest_api:documentfile-detail', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document.file_latest.pk
            }
        )

    def _request_test_document_file_download_api_view(self):
        return self.get(
            viewname='rest_api:documentfile-download', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document.file_latest.pk,
            }
        )

    def _request_test_document_file_list_api_view(self):
        return self.get(
            viewname='rest_api:documentfile-list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_file_upload_api_view(self):
        pk_list = list(DocumentFile.objects.values_list('pk', flat=True))

        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_descriptor:
            response = self.post(
                viewname='rest_api:documentfile-list', kwargs={
                    'document_id': self.test_document.pk,
                }, data={
                    'action': TEST_DOCUMENT_FILE_ACTION, 'comment': '',
                    'file_new': file_descriptor,
                }
            )

        try:
            self.test_document_file = DocumentFile.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentFile.DoesNotExist:
            self.test_document_file = None

        return response


class DocumentFileLinkTestMixin:
    def _resolve_test_document_file_link(self, test_link):
        self.add_test_view(test_object=self.test_document_file)
        context = self.get_test_view()
        return test_link.resolve(context=context)


class DocumentFileTestMixin:
    def _upload_new_file(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            self.test_document.file_new(
                comment=TEST_DOCUMENT_FILE_COMMENT, file_object=file_object
            )


class DocumentFileViewTestMixin:
    def _request_test_document_file_delete_view(self, document_file):
        return self.post(
            viewname='documents:document_file_delete', kwargs={
                'document_file_id': document_file.pk
            }
        )

    def _request_test_document_file_delete_multiple_view(self):
        return self.post(
            viewname='documents:document_file_delete_multiple', data={
                'id_list': self.test_document_file.pk
            }
        )

    def _request_test_document_file_download_view(self, data=None):
        data = data or {}
        return self.get(
            viewname='documents:document_file_download', kwargs={
                'document_file_id': self.test_document.file_latest.pk
            }, data=data
        )

    def _request_test_document_file_edit_view(self):
        return self.post(
            viewname='documents:document_file_edit', kwargs={
                'document_file_id': self.test_document_file.pk
            }, data={
                'comment': TEST_DOCUMENT_FILE_COMMENT_EDITED,
                'filename': TEST_DOCUMENT_FILE_FILENAME_EDITED
            }
        )

    def _request_test_document_file_list_view(self):
        return self.get(
            viewname='documents:document_file_list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_file_preview_view(self):
        return self.get(
            viewname='documents:document_file_preview', kwargs={
                'document_file_id': self.test_document_file.pk
            }
        )

    def _request_test_document_file_print_form_view(self):
        return self.get(
            viewname='documents:document_file_print_form', kwargs={
                'document_file_id': self.test_document_file.pk,
            }, data={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_file_print_view(self):
        return self.get(
            viewname='documents:document_file_print_view', kwargs={
                'document_file_id': self.test_document_file.pk,
            }, query={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_file_properties_view(self):
        return self.get(
            viewname='documents:document_file_properties', kwargs={
                'document_file_id': self.test_document_file.pk
            }
        )


class DocumentFilePageAPIViewTestMixin:
    def _request_test_document_file_page_detail_api_view(self):
        return self.get(
            viewname='rest_api:documentfilepage-detail', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document_file.pk,
                'document_file_page_id': self.test_document_file_page.pk
            }
        )

    def _request_test_document_file_page_image_api_view(
        self, maximum_layer_order=None
    ):
        return self.get(
            viewname='rest_api:documentfilepage-image', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document_file.pk,
                'document_file_page_id': self.test_document_file_page.pk
            }, query={'maximum_layer_order': maximum_layer_order}
        )

    def _request_test_document_file_page_list_api_view(self):
        return self.get(
            viewname='rest_api:documentfilepage-list', kwargs={
                'document_id': self.test_document.pk,
                'document_file_id': self.test_document_file.pk,
            }
        )


class DocumentFilePageViewTestMixin:
    def _request_test_document_file_page_count_update_view(self):
        return self.post(
            viewname='documents:document_file_page_count_update',
            kwargs={'document_file_id': self.test_document_file.pk}
        )

    def _request_test_document_file_multiple_page_count_update_view(self):
        return self.post(
            viewname='documents:document_file_multiple_page_count_update',
            data={'id_list': self.test_document_file.pk}
        )

    def _request_test_document_file_page_list_view(self):
        return self.get(
            viewname='documents:document_file_page_list', kwargs={
                'document_file_id': self.test_document_file.pk
            }
        )

    def _request_test_document_file_page_rotate_left_view(self):
        return self.post(
            viewname='documents:document_file_page_rotate_left', kwargs={
                'document_file_page_id': self.test_document_file_page.pk
            }
        )

    def _request_test_document_file_page_rotate_right_view(self):
        return self.post(
            viewname='documents:document_file_page_rotate_right', kwargs={
                'document_file_page_id': self.test_document_file_page.pk
            }
        )

    def _request_test_document_file_page_view(self, document_file_page):
        return self.get(
            viewname='documents:document_file_page_view', kwargs={
                'document_file_page_id': document_file_page.pk,
            }
        )

    def _request_test_document_file_page_zoom_in_view(self):
        return self.post(
            viewname='documents:document_file_page_zoom_in', kwargs={
                'document_file_page_id': self.test_document_file_page.pk
            }
        )

    def _request_test_document_file_page_zoom_out_view(self):
        return self.post(
            viewname='documents:document_file_page_zoom_out', kwargs={
                'document_file_page_id': self.test_document_file_page.pk
            }
        )


class DocumentFileTransformationTestMixin:
    def _create_document_file_transformation(self):
        layer_saved_transformations.add_transformation_to(
            obj=self.test_document_file.pages.first(),
            transformation_class=TEST_TRANSFORMATION_CLASS,
            arguments=TEST_TRANSFORMATION_ARGUMENT
        )


class DocumentFileTransformationViewTestMixin:
    def _request_test_document_file_transformations_clear_view(self):
        return self.post(
            viewname='documents:document_file_transformations_clear',
            kwargs={'document_file_id': self.test_document_file.pk}
        )

    def _request_test_document_file_multiple_transformations_clear_view(self):
        return self.post(
            viewname='documents:document_file_multiple_transformations_clear',
            data={'id_list': self.test_document_file.pk}
        )

    def _request_test_document_file_transformations_clone_view(self):
        return self.post(
            viewname='documents:document_file_transformations_clone',
            kwargs={'document_file_id': self.test_document_file.pk}, data={
                'page': self.test_document_file.pages.first().pk
            }
        )



# -*- coding: utf-8 -*-

import os

from django.conf import settings

from mayan.apps.common.literals import TIME_DELTA_UNIT_DAYS
from mayan.apps.converter.transformations import TransformationRotate

from ..literals import (
    DEFAULT_DOCUMENTS_RECENTLY_ACCESSED_COUNT,
    DEFAULT_DOCUMENTS_RECENTLY_CREATED_COUNT, DOCUMENT_FILE_ACTION_PAGES_NEW
)

DEFAULT_DOCUMENT_STUB_LABEL = 'document_stub'

# Filenames
TEST_COMPRESSED_DOCUMENTS_FILENAME = 'compressed_documents.zip'
TEST_DEU_DOCUMENT_FILENAME = 'deu_website.png'
TEST_DOCUMENT_DESCRIPTION = 'test description'
TEST_DOCUMENT_DESCRIPTION_EDITED = 'test document description edited'
TEST_DOCUMENT_FILE_COMMENT = 'test document file comment'
TEST_DOCUMENT_FILE_COMMENT_EDITED = 'test document file comment edited'
TEST_DOCUMENT_FILE_FILENAME_EDITED = 'test document file filename edited'
TEST_DOCUMENT_LABEL_EDITED = 'test document label edited'
TEST_DOCUMENT_TYPE_DELETE_PERIOD = 30
TEST_DOCUMENT_TYPE_DELETE_TIME_UNIT = TIME_DELTA_UNIT_DAYS
TEST_DOCUMENT_TYPE_LABEL = 'test_document_type'
TEST_DOCUMENT_TYPE_2_LABEL = 'test document type 2'
TEST_DOCUMENT_TYPE_LABEL_EDITED = 'test document type label edited'
TEST_DOCUMENT_TYPE_QUICK_LABEL = 'test quick label'
TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED = 'test quick label edited'
TEST_DOCUMENT_VERSION_COMMENT = 'test document version comment'
TEST_DOCUMENT_VERSION_COMMENT_EDITED = 'test document version comment edited'
TEST_DUPLICATED_DOCUMENT_LABEL = 'test duplicated document label'
TEST_HYBRID_DOCUMENT = 'hybrid_text_and_image.pdf'
TEST_MULTI_PAGE_TIFF = 'multi_page.tiff'
TEST_NON_ASCII_COMPRESSED_DOCUMENT_FILENAME = 'I18N_title_áéíóúüñÑ.png.zip'
TEST_NON_ASCII_DOCUMENT_FILENAME = 'I18N_title_áéíóúüñÑ.png'
TEST_OFFICE_DOCUMENT = 'simple_2_page_document.doc'
TEST_PDF_DOCUMENT_FILENAME = 'mayan_11_1.pdf'
TEST_SMALL_DOCUMENT_FILENAME = 'title_page.png'
TEST_SMALL_DOCUMENT_CHECKSUM = 'efa10e6cc21f83078aaa94d5cbe51de67b51af706143b\
afc7fd6d4c02124879a'
TEST_SMALL_DOCUMENT_MIMETYPE = 'image/png'
TEST_SMALL_DOCUMENT_SIZE = 17436
TEST_TRANSFORMATION_CLASS = TransformationRotate
TEST_TRANSFORMATION_NAME = 'rotate'
TEST_TRANSFORMATION_ARGUMENT = 'degrees: 180'

# File paths
TEST_COMPRESSED_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_COMPRESSED_DOCUMENTS_FILENAME
)
TEST_DEU_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_DEU_DOCUMENT_FILENAME
)
TEST_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_PDF_DOCUMENT_FILENAME
)
TEST_HYBRID_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_HYBRID_DOCUMENT
)
TEST_MULTI_PAGE_TIFF_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_MULTI_PAGE_TIFF
)
TEST_NON_ASCII_COMPRESSED_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_NON_ASCII_COMPRESSED_DOCUMENT_FILENAME
)
TEST_NON_ASCII_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_NON_ASCII_DOCUMENT_FILENAME
)
TEST_OFFICE_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_OFFICE_DOCUMENT
)
TEST_SMALL_DOCUMENT_PATH = os.path.join(
    settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
    'sample_documents', TEST_SMALL_DOCUMENT_FILENAME
)

# Test settings migrations

TEST_DOCUMENTS_CACHE_STORAGE_BACKEND = 'test.backend'
TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS = {'location': 'test value'}
TEST_DOCUMENTS_STORAGE_BACKEND = 'test.backend'
TEST_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS = {'location': 'test value'}
TEST_DOCUMENTS_RECENTLY_CREATED_COUNT = DEFAULT_DOCUMENTS_RECENTLY_CREATED_COUNT * 2
TEST_DOCUMENTS_RECENTLY_ACCESSED_COUNT = DEFAULT_DOCUMENTS_RECENTLY_ACCESSED_COUNT * 2

# Workflows

TEST_DOCUMENT_TYPE_CHANGE_ACTION_DOTTED_PATH = 'mayan.apps.documents.workflow_actions.DocumentTypeChangeAction'
TEST_TRASH_DOCUMENT_WORKFLOW_ACTION_DOTTED_PATH = 'mayan.apps.documents.workflow_actions.TrashDocumentAction'

# Others

TEST_DOCUMENT_FILE_ACTION = DOCUMENT_FILE_ACTION_PAGES_NEW



from django.db.models import Q

from ...classes import BaseDocumentFilenameGenerator
from ...models.document_type_models import DocumentType, DocumentTypeFilename

from ..literals import (
    TEST_DOCUMENT_TYPE_DELETE_PERIOD, TEST_DOCUMENT_TYPE_DELETE_TIME_UNIT,
    TEST_DOCUMENT_TYPE_LABEL, TEST_DOCUMENT_TYPE_LABEL_EDITED,
    TEST_DOCUMENT_TYPE_QUICK_LABEL, TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED
)


class DocumentQuickLabelViewTestMixin:
    def _request_test_document_quick_label_edit_view(self, extra_data=None):
        data = {
            'document_type_available_filenames': self.test_document_type_quick_label.pk,
            'label': ''
            # View needs at least an empty label for quick
            # label to work. Cause is unknown.
        }
        data.update(extra_data or {})

        return self.post(
            viewname='documents:document_properties_edit', kwargs={
                'document_id': self.test_document.pk
            }, data=data
        )


class DocumentTypeAPIViewTestMixin:
    def _request_test_document_type_create_api_view(self):
        pk_list = list(DocumentType.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='rest_api:documenttype-list', data={
                'label': TEST_DOCUMENT_TYPE_LABEL
            }
        )

        try:
            self.test_document_type = DocumentType.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentType.DoesNotExist:
            self.test_document_type = None

        return response

    def _request_test_document_type_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_test_document_type_detail_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_test_document_type_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={'label': TEST_DOCUMENT_TYPE_LABEL_EDITED}
        )

    def _request_test_document_type_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={'label': TEST_DOCUMENT_TYPE_LABEL_EDITED}
        )

    def _request_test_document_type_list_api_view(self):
        return self.get(viewname='rest_api:documenttype-list')


class DocumentTypeDeletionPoliciesViewTestMixin:
    def _request_test_document_type_policies_get_view(self):
        return self.get(
            viewname='documents:document_type_policies', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_test_document_type_policies_post_view(self):
        return self.post(
            viewname='documents:document_type_policies', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )


class DocumentTypeFilenameGeneratorViewTestMixin:
    def _request_test_document_type_filename_generator_get_view(self):
        return self.get(
            viewname='documents:document_type_filename_generator', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_test_document_type_filename_generator_post_view(self):
        return self.post(
            viewname='documents:document_type_filename_generator', kwargs={
                'document_type_id': self.test_document_type.pk
            }, data={
                'filename_generator_backend': BaseDocumentFilenameGenerator.get_default()
            }
        )


class DocumentTypeQuickLabelAPIViewTestMixin:
    def _request_test_document_type_quick_label_create_api_view(self):
        pk_list = list(DocumentTypeFilename.objects.values('pk'))

        response = self.post(
            viewname='rest_api:documenttype-quicklabel-list', kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL
            }
        )

        try:
            self.test_document_type_quick_label = DocumentTypeFilename.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentTypeFilename.DoesNotExist:
            self.test_document_type_quick_label = None

        return response

    def _request_test_document_type_quick_label_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
                'document_type_quick_label_id': self.test_document_type_quick_label.pk,
            }
        )

    def _request_test_document_type_quick_label_detail_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
                'document_type_quick_label_id': self.test_document_type_quick_label.pk,
            }
        )

    def _request_test_document_type_quick_label_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
                'document_type_quick_label_id': self.test_document_type_quick_label.pk,
            }, data={'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED}
        )

    def _request_test_document_type_quick_label_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self.test_document_type.pk,
                'document_type_quick_label_id': self.test_document_type_quick_label.pk,
            }, data={'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED}
        )

    def _request_test_document_type_quick_label_list_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-quicklabel-list', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )


class DocumentTypeQuickLabelViewTestMixin:
    def _request_test_quick_label_create_view(self):
        return self.post(
            viewname='documents:document_type_filename_create', kwargs={
                'document_type_id': self.test_document_type.pk
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL,
            }
        )

    def _request_test_quick_label_delete_view(self):
        return self.post(
            viewname='documents:document_type_filename_delete', kwargs={
                'document_type_filename_id': self.test_document_type_quick_label.pk
            }
        )

    def _request_test_quick_label_edit_view(self):
        return self.post(
            viewname='documents:document_type_filename_edit', kwargs={
                'document_type_filename_id': self.test_document_type_quick_label.pk
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED,
            }
        )

    def _request_test_quick_label_list_view(self):
        return self.get(
            viewname='documents:document_type_filename_list', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )


class DocumentTypeQuickLabelTestMixin:
    def _create_test_document_type_quick_label(self):
        self.test_document_type_quick_label = self.test_document_type.filenames.create(
            filename=TEST_DOCUMENT_TYPE_QUICK_LABEL
        )


class DocumentTypeViewTestMixin:
    def _request_test_document_type_create_view(self):
        return self.post(
            viewname='documents:document_type_create',
            data={
                'label': TEST_DOCUMENT_TYPE_LABEL,
                'delete_time_period': TEST_DOCUMENT_TYPE_DELETE_PERIOD,
                'delete_time_unit': TEST_DOCUMENT_TYPE_DELETE_TIME_UNIT
            }
        )

    def _request_test_document_type_delete_view(self):
        return self.post(
            viewname='documents:document_type_delete', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_test_document_type_edit_view(self):
        return self.post(
            viewname='documents:document_type_edit', kwargs={
                'document_type_id': self.test_document_type.pk
            }, data={
                'label': TEST_DOCUMENT_TYPE_LABEL_EDITED,
            }
        )

    def _request_test_document_type_list_view(self):
        return self.get(viewname='documents:document_type_list')



import logging

from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.views.generics import (
    ConfirmView, MultipleObjectConfirmActionView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectListView
)

from .forms import MessageCreateForm, MessageDetailForm
from .icons import icon_message_list
from .links import link_message_create
from .models import Message
from .permissions import (
    permission_message_create, permission_message_delete,
    permission_message_edit, permission_message_view
)

logger = logging.getLogger(name=__name__)


class MessageCreateView(SingleObjectCreateView):
    form_class = MessageCreateForm
    model = Message
    view_permission = permission_message_create

    def get_extra_context(self):
        return {
            'title': _('Create message')
        }

    def get_instance_extra_data(self):
        return {
            'sender_object': self.request.user,
            '_event_actor': self.request.user
        }


class MessageDeleteView(MultipleObjectConfirmActionView):
    error_message = _('Error deleting message "%(instance)s"; %(exception)s')
    object_permission = permission_message_delete
    pk_url_kwarg = 'message_id'
    post_action_redirect = reverse_lazy(viewname='messaging:message_list')
    success_message_single = _('Message "%(object)s" deleted successfully.')
    success_message_singular = _('%(count)d message deleted successfully.')
    success_message_plural = _('%(count)d messages deleted successfully.')
    title_single = _('Delete message: %(object)s.')
    title_singular = _('Delete the %(count)d selected message.')
    title_plural = _('Delete the %(count)d selected messages.')

    def get_extra_context(self):
        context = {
            'delete_view': True,
        }

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_source_queryset(self):
        return self.request.user.messages.all()

    def object_action(self, instance, form=None):
        instance.delete()


class MessageDetailView(SingleObjectDetailView):
    form_class = MessageDetailForm
    object_permission = permission_message_view
    pk_url_kwarg = 'message_id'

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)
        self.object._event_actor = self.request.user
        self.object.mark_read()
        return result

    def get_extra_context(self):
        return {
            'form_hide_help_text': True,
            'hide_labels': True,
            'object': self.object,
            'title': _('Details of message: %s') % self.object,
        }

    def get_initial(self):
        return {
            'body': self.object.get_rendered_body()
        }

    def get_source_queryset(self):
        return self.request.user.messages.all()


class MessageListView(SingleObjectListView):
    object_permission = permission_message_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_message_list,
            'no_results_main_link': link_message_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Here you will find text messages from other users or from '
                'the system.'
            ),
            'no_results_title': _('There are no messages'),
            'title': _('Messages'),
        }

    def get_source_queryset(self):
        return self.request.user.messages.all()


class MessageMarkReadView(MultipleObjectConfirmActionView):
    error_message = _(
        'Error marking message "%(instance)s" as read; %(exception)s'
    )
    object_permission = permission_message_edit
    pk_url_kwarg = 'message_id'
    post_action_redirect = reverse_lazy(viewname='messaging:message_list')
    success_message_single = _(
        'Message "%(object)s" marked as read successfully.'
    )
    success_message_singular = _(
        '%(count)d message marked as read successfully.'
    )
    success_message_plural = _(
        '%(count)d messages marked as read successfully.'
    )
    title_single = _('Mark the message "%(object)s" as read.')
    title_singular = _('Mark the %(count)d selected message as read.')
    title_plural = _('Mark the %(count)d selected messages as read.')

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_source_queryset(self):
        return self.request.user.messages.all()

    def object_action(self, instance, form=None):
        instance._event_actor = self.request.user
        instance.mark_read()


class MessageMarkReadAllView(ConfirmView):
    post_action_redirect = reverse_lazy(viewname='messaging:message_list')

    def get_extra_context(self):
        return {
            'title': _('Mark all message as read?')
        }

    def get_queryset(self):
        return self.request.user.messages.all()

    def view_action(self, form=None):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_message_edit, queryset=self.get_queryset(),
            user=self.request.user
        )

        for message in queryset.all():
            message._event_actor = self.request.user
            message.mark_read()

        messages.success(
            message=_('All messages marked as read.'),
            request=self.request
        )


class MessageMarkUnReadView(MultipleObjectConfirmActionView):
    error_message = _(
        'Error marking message "%(instance)s" as unread; %(exception)s'
    )
    object_permission = permission_message_edit
    pk_url_kwarg = 'message_id'
    post_action_redirect = reverse_lazy(viewname='messaging:message_list')
    success_message_single = _(
        'Message "%(object)s" marked as unread successfully.'
    )
    success_message_singular = _(
        '%(count)d message marked as unread successfully.'
    )
    success_message_plural = _(
        '%(count)d messages marked as unread successfully.'
    )
    title_single = _('Mark the message "%(object)s" as unread.')
    title_singular = _('Mark the %(count)d selected message as unread.')
    title_plural = _('Mark the %(count)d selected messages as unread.')

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_source_queryset(self):
        return self.request.user.messages.all()

    def object_action(self, instance, form=None):
        instance._event_actor = self.request.user
        instance.mark_unread()



from pathlib import Path
import shutil

from mayan.apps.documents.models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import (
    TEST_NON_ASCII_DOCUMENT_FILENAME,
    TEST_NON_ASCII_COMPRESSED_DOCUMENT_PATH, TEST_SMALL_DOCUMENT_CHECKSUM,
    TEST_SMALL_DOCUMENT_PATH
)

from ..source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_ALWAYS

from .literals import TEST_WATCHFOLDER_SUBFOLDER
from .mixins.watch_folder_source_mixins import WatchFolderSourceTestMixin


class WatchFolderSourceBackendTestCase(
    WatchFolderSourceTestMixin, GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_exclude_regular_expression(self):
        path = Path(TEST_SMALL_DOCUMENT_PATH)

        self._create_test_watch_folder(
            extra_data={'exclude_regex': path.name}
        )

        document_count = Document.objects.count()

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        shutil.copy(src=TEST_SMALL_DOCUMENT_PATH, dst=temporary_directory)

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count)

        backend_data = self.test_source.get_backend_data()
        backend_data['exclude_regex'] = ''
        self.test_source.set_backend_data(obj=backend_data)

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count + 1)

    def test_include_regular_expression(self):
        path = Path(TEST_SMALL_DOCUMENT_PATH)

        self._create_test_watch_folder(
            extra_data={'include_regex': '_____.*'}
        )

        document_count = Document.objects.count()

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        shutil.copy(src=TEST_SMALL_DOCUMENT_PATH, dst=temporary_directory)

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count)

        backend_data = self.test_source.get_backend_data()
        backend_data['include_regex'] = path.name
        self.test_source.set_backend_data(obj=backend_data)

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count + 1)

    def test_upload_simple_file(self):
        self._create_test_watch_folder()

        document_count = Document.objects.count()

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        shutil.copy(src=TEST_SMALL_DOCUMENT_PATH, dst=temporary_directory)

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )

    def test_subfolder_disabled(self):
        self._create_test_watch_folder()

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        test_path = Path(temporary_directory)
        test_subfolder = test_path.joinpath(TEST_WATCHFOLDER_SUBFOLDER)
        test_subfolder.mkdir()

        shutil.copy(
            src=TEST_SMALL_DOCUMENT_PATH, dst=test_subfolder
        )

        document_count = Document.objects.count()

        self.test_source.get_backend_instance().process_documents()
        self.assertEqual(Document.objects.count(), document_count)

    def test_subfolder_enabled(self):
        self._create_test_watch_folder(
            extra_data={'include_subdirectories': True}
        )

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        test_path = Path(temporary_directory)
        test_subfolder = test_path.joinpath(TEST_WATCHFOLDER_SUBFOLDER)
        test_subfolder.mkdir()

        shutil.copy(src=TEST_SMALL_DOCUMENT_PATH, dst=test_subfolder)

        document_count = Document.objects.count()

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count + 1)

        document = Document.objects.first()

        self.assertEqual(
            document.file_latest.checksum, TEST_SMALL_DOCUMENT_CHECKSUM
        )

    def test_non_ascii_file_in_non_ascii_compressed_file(self):
        """
        Test Non-ASCII named documents inside Non-ASCII named compressed
        file. GitHub issue #163.
        """
        self._create_test_watch_folder(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        temporary_directory = self.test_source.get_backend_data()['folder_path']

        shutil.copy(
            src=TEST_NON_ASCII_COMPRESSED_DOCUMENT_PATH,
            dst=temporary_directory
        )

        document_count = Document.objects.count()

        self.test_source.get_backend_instance().process_documents()

        self.assertEqual(Document.objects.count(), document_count + 1)

        document = Document.objects.first()

        self.assertEqual(document.label, TEST_NON_ASCII_DOCUMENT_FILENAME)
        self.assertEqual(document.file_latest.exists(), True)
        self.assertEqual(document.file_latest.size, 17436)
        self.assertEqual(document.file_latest.mimetype, 'image/png')
        self.assertEqual(document.file_latest.encoding, 'binary')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Converter'), name='converter')

event_asset_created = namespace.add_event_type(
    label=_('Asset created'), name='asset_created'
)
event_asset_edited = namespace.add_event_type(
    label=_('Asset edited'), name='asset_edited'
)



from mayan.apps.appearance.classes import Icon

icon_dashboard_detail = Icon(
    driver_name='fontawesome', symbol='tachometer-alt'
)
icon_dashboard_link_icon = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)
icon_dashboard_list = Icon(
    driver_name='fontawesome', symbol='tachometer-alt'
)



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from .mixins import DocumentCommentTestMixin


class DocumentCommentModelTestCase(
    DocumentCommentTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_user()
        self._create_test_document_stub()

    def test_method_get_absolute_url(self):
        self._create_test_comment()

        self.assertTrue(self.test_document_comment.get_absolute_url())



from unittest import skip

from mayan.apps.documents.models.document_models import DocumentSearchResult
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import document_search
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..classes import SearchModel
from ..permissions import permission_search_tools

from .mixins import (
    SearchTestMixin, SearchToolsViewTestMixin, SearchViewTestMixin
)


class AdvancedSearchViewTestCaseMixin(
    DocumentTestMixin, SearchViewTestMixin
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self.test_document_count = 4

        # Upload many instances of the same test document.
        for i in range(self.test_document_count):
            self._upload_test_document()
            self.grant_access(
                obj=self.test_document, permission=permission_document_view
            )

    def test_advanced_search_past_first_page(self):
        test_document_label = self.test_documents[0].label

        # Make sure all documents are returned by the search
        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': test_document_label},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), self.test_document_count)

        with self.settings(VIEWS_PAGINATE_BY=2):
            # Functional test for the first page of advanced results
            response = self._request_search_results_view(
                data={'label': test_document_label}, kwargs={
                    'search_model_name': document_search.get_full_name()
                }
            )

            # Total (1 - 2 out of 4) (Page 1 of 2)
            # 4 results total, 2 pages, current page is 1,
            # object in this page: 2
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.context['paginator'].object_list.count(), 4
            )
            self.assertEqual(response.context['paginator'].num_pages, 2)
            self.assertEqual(response.context['page_obj'].number, 1)
            self.assertEqual(
                response.context['page_obj'].object_list.count(), 2
            )

            # Functional test for the second page of advanced results
            response = self._request_search_results_view(
                data={'label': test_document_label, 'page': 2}, kwargs={
                    'search_model_name': document_search.get_full_name()
                }
            )
            # Total (3 - 4 out of 4) (Page 2 of 2)
            # 4 results total, 2 pages, current page is 1,
            # object in this page: 2
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.context['paginator'].object_list.count(), 4
            )
            self.assertEqual(response.context['paginator'].num_pages, 2)
            self.assertEqual(response.context['page_obj'].number, 2)
            self.assertEqual(
                response.context['page_obj'].object_list.count(), 2
            )


class DjangoAdvancedSearchViewTestCase(
    AdvancedSearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchAdvancedSearchViewTestCase(
    AdvancedSearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshAdvancedSearchViewTestCase(
    AdvancedSearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""


class SearchViewTestCaseMixin(DocumentTestMixin, SearchViewTestMixin):
    def test_result_view_with_search_mode_in_data(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_results_view(
            data={
                'label': self.test_document.label,
                '_search_model_name': document_search.get_full_name()
            }
        )
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )


class DjangoSearchViewTestCase(
    SearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchSearchViewTestCase(
    SearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchViewTestCase(
    SearchViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""


class SearchToolsViewTestCaseMixin(
    DocumentTestMixin, SearchToolsViewTestMixin, SearchTestMixin
):
    def setUp(self):
        super().setUp()

        self.document_search_model = SearchModel.get_for_model(
            instance=DocumentSearchResult
        )

    def test_search_backend_reindex_view_no_permission(self):
        self.search_backend.reset(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 403)

    def test_search_backend_reindex_view_artifacts_no_permission(self):
        self.search_backend.reset(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 403)

        queryset = self.search_backend.search(
            search_model=self.document_search_model,
            query={'q': self.test_document.label},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_search_backend_reindex_view_with_permission(self):
        self.search_backend.reset(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_permission(permission=permission_search_tools)

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 302)

    def test_search_backend_reindex_view_artifacts_with_permission(self):
        self.search_backend.reset(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_permission(permission=permission_search_tools)

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 302)

        queryset = self.search_backend.search(
            search_model=self.document_search_model,
            query={'q': self.test_document.label},
            user=self._test_case_user
        )
        self.assertNotEqual(queryset.count(), 0)


class DjangoSearchToolViewTestCase(
    SearchToolsViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""

    @skip('Backend does not support indexing.')
    def test_search_backend_reindex_view_artifacts_no_permission(self):
        """Backend does not support indexing."""

    @skip('Backend does not support indexing.')
    def test_search_backend_reindex_view_artifacts_with_permission(self):
        """Backend does not support indexing."""


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchToolViewTestCase(
    SearchToolsViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchToolViewTestCase(
    SearchToolsViewTestCaseMixin, GenericViewTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('storage', '0004_downloadfile'),
    ]

    operations = [
        migrations.AlterField(
            model_name='shareduploadedfile',
            name='filename',
            field=models.CharField(db_index=True, max_length=255, verbose_name='Filename'),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0075_delete_duplicateddocumentold'),
        ('document_indexing', '0020_auto_20210307_0405'),
    ]

    operations = [
        migrations.DeleteModel(
            name='IndexInstance',
        ),
        migrations.RenameModel(
            old_name='Index',
            new_name='IndexTemplate',
        ),
    ]



def load_env_file(filename='config.env'):
    result = {}
    with open(file=filename) as file_object:
        for line in file_object:
            if not line.startswith('#'):
                key, value = line.strip().split('=')

                result[key] = value

    return result



from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.classes import Layer

from .permissions import (
    permission_redaction_create, permission_redaction_delete,
    permission_redaction_edit, permission_redaction_exclude,
    permission_redaction_view
)

layer_redactions = Layer(
    empty_results_text=_(
        'Redactions allow removing access to confidential and '
        'sensitive information without having to modify the document.'
    ), label=_('Redactions'), name='redactions',
    order=0, permissions={
        'create': permission_redaction_create,
        'delete': permission_redaction_delete,
        'exclude': permission_redaction_exclude,
        'edit': permission_redaction_edit,
        'select': permission_redaction_create,
        'view': permission_redaction_view,
    }, symbol='highlighter'
)



from django.apps import apps
from django.utils.encoding import force_text
from django.utils.html import mark_safe, escape

from .icons import (
    icon_index, icon_index_level_up, icon_index_instance_node_with_documents
)


def get_instance_link(index_instance_node):
    """
    Return an HTML anchor to an index node instance
    """
    return mark_safe(
        s='<a href="{url}">{text}</a>'.format(
            url=index_instance_node.get_absolute_url(),
            text=escape(index_instance_node.get_full_path())
        )
    )


def index_instance_item_link(index_instance_item):
    IndexInstanceNode = apps.get_model(
        app_label='document_indexing', model_name='IndexInstanceNode'
    )

    if isinstance(index_instance_item, IndexInstanceNode):
        if index_instance_item.index_template_node.link_documents:
            icon = icon_index_instance_node_with_documents.render()
        else:
            icon = icon_index_level_up.render()
    else:
        icon = ''

    return mark_safe(
        s='%(icon)s&nbsp;<a href="%(url)s">%(text)s</a>' % {
            'url': index_instance_item.get_absolute_url(),
            'icon': icon,
            'text': index_instance_item
        }
    )


def node_level(node):
    """
    Render an indented tree like output for a specific node
    """
    return mark_safe(
        s=''.join(
            [
                '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;' * node.get_level(),
                '' if node.is_root_node() else icon_index_level_up.render(),
                force_text(s=node)
            ]
        )
    )


def node_tree(node, user):
    result = []

    result.append('<div class="list-group">')

    for ancestor in node.get_ancestors(include_self=True):
        if ancestor.is_root_node():
            element = node.index()
            icon = icon_index
        else:
            element = ancestor
            if element.index_template_node.link_documents:
                icon = icon_index_instance_node_with_documents
            else:
                icon = icon_index_level_up

        result.append(
            '<a href="{url}" class="list-group-item {active}"><span class="badge">{count}</span>{icon} {text}</a>'.format(
                url=element.get_absolute_url(),
                active='active' if element == node or node.get_ancestors(include_self=True).count() == 1 else '',
                count=element.get_descendants_document_count(user=user),
                icon=icon.render(),
                text=escape(element)
            )
        )

    result.append('</div>')

    return mark_safe(''.join(result))



import json

from django.db.models import Q

from ..models import UserMailer

from .literals import (
    TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS,
    TEST_USER_MAILER_BACKEND_PATH, TEST_USER_MAILER_LABEL
)


class DocumentMailerViewTestMixin:
    def _request_test_document_send_link_single_view(self):
        return self.post(
            viewname='mailer:send_document_link_single', kwargs={
                'document_id': self.test_document.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_send_link_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_link_multiple', query={
                'id_list': self.test_document.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )


class DocumentFileMailerViewTestMixin:
    def _request_test_document_file_send_link_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_single', kwargs={
                'document_file_id': self.test_document_file.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_send_link_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_multiple', query={
                'id_list': self.test_document_file.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_single', kwargs={
                'document_file_id': self.test_document_file.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_multiple', query={
                'id_list': self.test_document_file.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )


class DocumentVersionMailerViewTestMixin:
    def _request_test_document_version_send_link_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_single', kwargs={
                'document_version_id': self.test_document_version.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_version_send_link_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_multiple', query={
                'id_list': self.test_document_version.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_version_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_single',
            kwargs={
                'document_version_id': self.test_document_version.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_version_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_multiple',
            query={
                'id_list': self.test_document_version.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )


class MailerTestMixin:
    def _create_test_user_mailer(self):
        self._test_user_mailer = UserMailer.objects.create(
            default=True,
            enabled=True,
            label=TEST_USER_MAILER_LABEL,
            backend_path=TEST_USER_MAILER_BACKEND_PATH,
            backend_data=json.dumps(
                obj={
                    'from': TEST_EMAIL_FROM_ADDRESS
                }
            )
        )


class MailerViewTestMixin:
    def _request_test_user_mailer_create_view(self):
        pk_list = list(UserMailer.objects.values('pk'))

        response = self.post(
            viewname='mailer:user_mailer_create', kwargs={
                'class_path': TEST_USER_MAILER_BACKEND_PATH
            }, data={
                'default': True,
                'enabled': True,
                'label': TEST_USER_MAILER_LABEL,
            }
        )

        try:
            self._test_user_mailer = UserMailer.objects.get(
                ~Q(pk__in=pk_list)
            )
        except UserMailer.DoesNotExist:
            self._test_user_mailer = None

        return response

    def _request_test_user_mailer_delete_view(self):
        return self.post(
            viewname='mailer:user_mailer_delete', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }
        )

    def _request_test_user_mailer_edit_view(self):
        return self.post(
            viewname='mailer:user_mailer_edit', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }, data={
                'label': '{}_edited'.format(TEST_USER_MAILER_LABEL)
            }
        )

    def _request_test_user_mailer_list_view(self):
        return self.get(
            viewname='mailer:user_mailer_list'
        )

    def _request_test_user_mailer_log_entry_view(self):
        return self.get(
            viewname='mailer:user_mailer_log', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }
        )

    def _request_test_user_mailer_test_view(self):
        return self.post(
            viewname='mailer:user_mailer_test', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }, data={
                'email': getattr(
                    self, 'test_email_address', TEST_EMAIL_ADDRESS
                )
            }
        )



from django.contrib import admin

from .models import SmartLink, SmartLinkCondition


class SmartLinkConditionInline(admin.StackedInline):
    model = SmartLinkCondition
    extra = 1
    classes = ('collapse-open',)
    allow_add = True


@admin.register(SmartLink)
class SmartLinkAdmin(admin.ModelAdmin):
    def document_type_list(self, instance):
        return ','.join(
            instance.document_types.values_list('label', flat=True)
        )

    filter_horizontal = ('document_types',)
    inlines = (SmartLinkConditionInline,)
    list_display = ('label', 'dynamic_label', 'enabled', 'document_type_list')
